import torch
from torch import nn as nn
from torch.nn import functional as F


def conv3d(in_channels, out_channels, kernel_size, bias, padding=1):
    return nn.Conv3d(in_channels, out_channels, kernel_size, padding=padding, bias=bias)


def create_conv(in_channels, out_channels, kernel_size, order, num_groups, padding=1):
    """
    Create a list of modules with together constitute a single conv layer with non-linearity
    and optional batchnorm/groupnorm.

    Args:
        in_channels (int): number of input channels
        out_channels (int): number of output channels
        order (string): order of things, e.g.
            'cr' -> conv + ReLU
            'crg' -> conv + ReLU + groupnorm
            'cl' -> conv + LeakyReLU
            'ce' -> conv + ELU
        num_groups (int): number of groups for the GroupNorm
        padding (int): add zero-padding to the input

    Return:
        list of tuple (name, module)
    """
    assert 'c' in order, "Conv layer MUST be present"
    assert order[0] not in 'rle', 'Non-linearity cannot be the first operation in the layer'

    modules = []
    for i, char in enumerate(order):
        if char == 'r':
            modules.append(('ReLU', nn.ReLU(inplace=True)))
        elif char == 'l':
            modules.append(('LeakyReLU', nn.LeakyReLU(negative_slope=0.1, inplace=True)))
        elif char == 'e':
            modules.append(('ELU', nn.ELU(inplace=True)))
        elif char == 'c':
            # add learnable bias only in the absence of gatchnorm/groupnorm
            bias = not ('g' in order or 'b' in order)
            modules.append(('conv', conv3d(in_channels, out_channels, kernel_size, bias, padding=padding)))
        elif char == 'g':
            is_before_conv = i < order.index('c')
            assert not is_before_conv, 'GroupNorm MUST go after the Conv3d'
            # number of groups must be less or equal the number of channels
            if out_channels < num_groups:
                num_groups = out_channels
            modules.append(('groupnorm', nn.GroupNorm(num_groups=num_groups, num_channels=out_channels)))
        elif char == 'b':
            is_before_conv = i < order.index('c')
            if is_before_conv:
                modules.append(('batchnorm', nn.BatchNorm3d(in_channels)))
            else:
                modules.append(('batchnorm', nn.BatchNorm3d(out_channels)))
        else:
            raise ValueError(f"Unsupported layer type '{char}'. MUST be one of ['b', 'g', 'r', 'l', 'e', 'c']")

    return modules


class SingleConv(nn.Sequential):
    """
    Basic convolutional module consisting of a Conv3d, non-linearity and optional batchnorm/groupnorm. The order
    of operations can be specified via the `order` parameter

    Args:
        in_channels (int): number of input channels
        out_channels (int): number of output channels
        kernel_size (int): size of the convolving kernel
        order (string): determines the order of layers, e.g.
            'cr' -> conv + ReLU
            'crg' -> conv + ReLU + groupnorm
            'cl' -> conv + LeakyReLU
            'ce' -> conv + ELU
        num_groups (int): number of groups for the GroupNorm
    """

    def __init__(self, in_channels, out_channels, kernel_size=3, order='cr', num_groups=8, padding=1):
        super(SingleConv, self).__init__()

        for name, module in create_conv(in_channels, out_channels, kernel_size, order, num_groups, padding=padding):
            self.add_module(name, module)


class DoubleConv(nn.Sequential):
    """
    A module consisting of two consecutive convolution layers (e.g. BatchNorm3d+ReLU+Conv3d).
    We use (Conv3d+ReLU+GroupNorm3d) by default.
    This can be changed however by providing the 'order' argument, e.g. in order
    to change to Conv3d+BatchNorm3d+ELU use order='cbe'.
    Use padded convolutions to make sure that the output (H_out, W_out) is the same
    as (H_in, W_in), so that you don't have to crop in the decoder path.

    Args:
        in_channels (int): number of input channels
        out_channels (int): number of output channels
        encoder (bool): if True we're in the encoder path, otherwise we're in the decoder
        kernel_size (int): size of the convolving kernel
        order (string): determines the order of layers, e.g.
            'cr' -> conv + ReLU
            'crg' -> conv + ReLU + groupnorm
            'cl' -> conv + LeakyReLU
            'ce' -> conv + ELU
        num_groups (int): number of groups for the GroupNorm
    """

    def __init__(self, in_channels, out_channels, encoder, kernel_size=3, order='cr', num_groups=8):
        super(DoubleConv, self).__init__()
        if encoder:
            # we're in the encoder path
            conv1_in_channels = in_channels
            conv1_out_channels = out_channels // 2
            if conv1_out_channels < in_channels:
                conv1_out_channels = in_channels
            conv2_in_channels, conv2_out_channels = conv1_out_channels, out_channels
        else:
            # we're in the decoder path, decrease the number of channels in the 1st convolution
            conv1_in_channels, conv1_out_channels = in_channels, out_channels
            conv2_in_channels, conv2_out_channels = out_channels, out_channels

        # conv1
        self.add_module('SingleConv1',
                        SingleConv(conv1_in_channels, conv1_out_channels, kernel_size, order, num_groups))
        # conv2
        self.add_module('SingleConv2',
                        SingleConv(conv2_in_channels, conv2_out_channels, kernel_size, order, num_groups))


class ExtResNetBlock(nn.Module):
    """
    Basic UNet block consisting of a SingleConv followed by the residual block.
    The SingleConv takes care of increasing/decreasing the number of channels and also ensures that the number
    of output channels is compatible with the residual block that follows.
    This block can be used instead of standard DoubleConv in the Encoder module.
    Motivated by: https://arxiv.org/pdf/1706.00120.pdf

    Notice we use ELU instead of ReLU (order='cge') and put non-linearity after the groupnorm.
    """

    def __init__(self, in_channels, out_channels, kernel_size=3, order='cge', num_groups=8, **kwargs):
        super(ExtResNetBlock, self).__init__()

        # first convolution
        self.conv1 = SingleConv(in_channels, out_channels, kernel_size=kernel_size, order=order, num_groups=num_groups)
        # residual block
        self.conv2 = SingleConv(out_channels, out_channels, kernel_size=kernel_size, order=order, num_groups=num_groups)
        # remove non-linearity from the 3rd convolution since it's going to be applied after adding the residual
        n_order = order
        for c in 'rel':
            n_order = n_order.replace(c, '')
        self.conv3 = SingleConv(out_channels, out_channels, kernel_size=kernel_size, order=n_order,
                                num_groups=num_groups)

        # create non-linearity separately
        if 'l' in order:
            self.non_linearity = nn.LeakyReLU(negative_slope=0.1, inplace=True)
        elif 'e' in order:
            self.non_linearity = nn.ELU(inplace=True)
        else:
            self.non_linearity = nn.ReLU(inplace=True)

    def forward(self, x):
        # apply first convolution and save the output as a residual
        out = self.conv1(x)
        residual = out

        # residual block
        out = self.conv2(out)
        out = self.conv3(out)

        out += residual
        out = self.non_linearity(out)

        return out


class Encoder(nn.Module):
    """
    A single module from the encoder path consisting of the optional max
    pooling layer (one may specify the MaxPool kernel_size to be different
    than the standard (2,2,2), e.g. if the volumetric data is anisotropic
    (make sure to use complementary scale_factor in the decoder path) followed by
    a DoubleConv module.
    Args:
        in_channels (int): number of input channels
        out_channels (int): number of output channels
        conv_kernel_size (int): size of the convolving kernel
        apply_pooling (bool): if True use MaxPool3d before DoubleConv
        pool_kernel_size (tuple): the size of the window to take a max over
        pool_type (str): pooling layer: 'max' or 'avg'
        basic_module(nn.Module): either ResNetBlock or DoubleConv
        conv_layer_order (string): determines the order of layers
            in `DoubleConv` module. See `DoubleConv` for more info.
        num_groups (int): number of groups for the GroupNorm
    """

    def __init__(self, in_channels, out_channels, conv_kernel_size=3, apply_pooling=True,
                 pool_kernel_size=(2, 2, 2), pool_type='max', basic_module=DoubleConv, conv_layer_order='cr',
                 num_groups=8):
        super(Encoder, self).__init__()
        assert pool_type in ['max', 'avg']
        if apply_pooling:
            if pool_type == 'max':
                self.pooling = nn.MaxPool3d(kernel_size=pool_kernel_size)
            else:
                self.pooling = nn.AvgPool3d(kernel_size=pool_kernel_size)
        else:
            self.pooling = None

        self.basic_module = basic_module(in_channels, out_channels,
                                         encoder=True,
                                         kernel_size=conv_kernel_size,
                                         order=conv_layer_order,
                                         num_groups=num_groups)

    def forward(self, x):
        if self.pooling is not None:
            x = self.pooling(x)
        x = self.basic_module(x)
        return x


class Decoder(nn.Module):
    """
    A single module for decoder path consisting of the upsample layer
    (either learned ConvTranspose3d or interpolation) followed by a DoubleConv
    module.
    Args:
        in_channels (int): number of input channels
        out_channels (int): number of output channels
        kernel_size (int): size of the convolving kernel
        scale_factor (tuple): used as the multiplier for the image H/W/D in
            case of nn.Upsample or as stride in case of ConvTranspose3d, must reverse the MaxPool3d operation
            from the corresponding encoder
        basic_module(nn.Module): either ResNetBlock or DoubleConv
        conv_layer_order (string): determines the order of layers
            in `DoubleConv` module. See `DoubleConv` for more info.
        num_groups (int): number of groups for the GroupNorm
    """

    def __init__(self, in_channels, out_channels, kernel_size=3,
                 scale_factor=(2, 2, 2), basic_module=DoubleConv, conv_layer_order='cr', num_groups=8):
        super(Decoder, self).__init__()
        if basic_module == DoubleConv:
            # if DoubleConv is the basic_module use nearest neighbor interpolation for upsampling
            self.upsample = None
        else:
            # otherwise use ConvTranspose3d (bear in mind your GPU memory)
            # make sure that the output size reverses the MaxPool3d from the corresponding encoder
            # (D_out = (D_in − 1) ×  stride[0] − 2 ×  padding[0] +  kernel_size[0] +  output_padding[0])
            # also scale the number of channels from in_channels to out_channels so that summation joining
            # works correctly
            self.upsample = nn.ConvTranspose3d(in_channels,
                                               out_channels,
                                               kernel_size=kernel_size,
                                               stride=scale_factor,
                                               padding=1,
                                               output_padding=1)
            # adapt the number of in_channels for the ExtResNetBlock
            in_channels = out_channels

        self.basic_module = basic_module(in_channels, out_channels,
                                         encoder=False,
                                         kernel_size=kernel_size,
                                         order=conv_layer_order,
                                         num_groups=num_groups)

    def forward(self, encoder_features, x):
        if self.upsample is None:
            # use nearest neighbor interpolation and concatenation joining
            output_size = encoder_features.size()[2:]
            x = F.interpolate(x, size=output_size, mode='nearest')
            # concatenate encoder_features (encoder path) with the upsampled input across channel dimension
            x = torch.cat((encoder_features, x), dim=1)
        else:
            # use ConvTranspose3d and summation joining
            x = self.upsample(x)
            x += encoder_features

        x = self.basic_module(x)
        return x


class FinalConv(nn.Sequential):
    """
    A module consisting of a convolution layer (e.g. Conv3d+ReLU+GroupNorm3d) and the final 1x1 convolution
    which reduces the number of channels to 'out_channels'.
    with the number of output channels 'out_channels // 2' and 'out_channels' respectively.
    We use (Conv3d+ReLU+GroupNorm3d) by default.
    This can be change however by providing the 'order' argument, e.g. in order
    to change to Conv3d+BatchNorm3d+ReLU use order='cbr'.
    Args:
        in_channels (int): number of input channels
        out_channels (int): number of output channels
        kernel_size (int): size of the convolving kernel
        order (string): determines the order of layers, e.g.
            'cr' -> conv + ReLU
            'crg' -> conv + ReLU + groupnorm
        num_groups (int): number of groups for the GroupNorm
    """

    def __init__(self, in_channels, out_channels, kernel_size=3, order='cr', num_groups=8):
        super(FinalConv, self).__init__()

        # conv1
        self.add_module('SingleConv', SingleConv(in_channels, in_channels, kernel_size, order, num_groups))

        # in the last layer a 1×1 convolution reduces the number of output channels to out_channels
        final_conv = nn.Conv3d(in_channels, out_channels, 1)
        self.add_module('final_conv', final_conv)



import os
import torch
import torch.nn as nn
from torch.autograd import Variable
from torch.utils.data import DataLoader
import argparse
import time
import datetime
import sys
import math
import scipy.io as scio
from skimage import io
import numpy as np
import math

from deepwonder.SEG.network import SEG_Network_3D_Unet
from deepwonder.SEG.data_process import test_preprocess_lessMemoryNoTail_SubImgSEG, testset, singlebatch_test_save, multibatch_test_save
from deepwonder.SEG.utils import FFDrealign4, inv_FFDrealign4

def seg_3dunet_ffd(net,
                sub_img,
                SEG_ffd = True,
                SEG_GPU='0', 
                SEG_batch_size=1, 
                SEG_img_w=256, 
                SEG_img_h=256, 
                SEG_img_s=64, 
                SEG_gap_w=224, 
                SEG_gap_h=224, 
                SEG_gap_s=32, 
                SEG_normalize_factor=1000):
    #############################################################################################################################################
    opt = {}
    #############################################################################################################################################
    opt['GPU'] = SEG_GPU
    opt['batch_size'] = SEG_batch_size
    opt['img_w'] = SEG_img_w
    opt['img_h'] = SEG_img_h
    opt['img_s'] = SEG_img_s

    opt['gap_w'] = SEG_gap_w
    opt['gap_h'] = SEG_gap_h
    opt['gap_s'] = SEG_gap_s

    opt['normalize_factor'] = SEG_normalize_factor
    #############################################################################################################################################
    os.environ["CUDA_VISIBLE_DEVICES"] = str(opt['GPU'])
    ##############################################################################################################################################################
    if torch.cuda.is_available():
        net = net.cuda()
    cuda = True if torch.cuda.is_available() else False
    Tensor = torch.cuda.FloatTensor if cuda else torch.FloatTensor
    ##############################################################################################################################################################
    name_list, noise_img, coordinate_list = test_preprocess_lessMemoryNoTail_SubImgSEG(opt, sub_img)

        #print(len(name_list))
    prev_time = time.time()
    time_start = time.time()

    num_s = math.ceil((noise_img.shape[0]-SEG_img_s+SEG_gap_s)/SEG_gap_s)
    denoise_img = np.zeros((num_s, noise_img.shape[1], noise_img.shape[2]))

    test_data = testset(name_list, coordinate_list, noise_img)
    testloader = DataLoader(test_data, batch_size=opt['batch_size'], shuffle=False, num_workers=4)
    for iteration, (noise_patch,single_coordinate) in enumerate(testloader):
        noise_patch=noise_patch.cuda().float()

        if SEG_ffd:
            real_A = FFDrealign4(noise_patch)
        if not SEG_ffd:
            real_A = noise_patch
        real_A = Variable(real_A)
                    
        fake_B = net(real_A)
        ################################################################################################################
        # Determine approximate time left
        batches_done = iteration
        batches_left = 1 * len(testloader) - batches_done
        time_left_seconds = int(batches_left * (time.time() - prev_time))
        time_left = datetime.timedelta(seconds=time_left_seconds)
        prev_time = time.time()
        ################################################################################################################
        if iteration % 1 == 0:
            time_end = time.time()
            time_cost = time_end - time_start  # datetime.timedelta(seconds= (time_end - time_start))
            # printf('\033[1;40;32m color!!! \033[0m hello\n')
            # print('\033[1;31mUsing GPU for training -----> \033[0m')
            print(
                '\r\033[1;31m[SEG]\033[0m [Patch %d/%d] [Time Cost: %.0d s] [ETA: %s s]    '
                % ( iteration + 1,
                    len(testloader),
                    time_cost,
                    time_left_seconds
                ), end=' ')
            # print(noise_patch.cpu().detach().numpy().max(),' ---> ', noise_patch.cpu().detach().numpy().min())

        if (iteration + 1) % len(testloader) == 0:
            print('\n', end=' ')
        ################################################################################################################
        if SEG_ffd:
            fake_B_realign = inv_FFDrealign4(fake_B)
        if not SEG_ffd:
            fake_B_realign = fake_B

        # fake_B_realign = fake_B
        output_image = np.squeeze(fake_B_realign.cpu().detach().numpy())
        # real_A_realign = inv_FFDrealign4(real_A)
        real_A_realign = real_A
        raw_image = np.squeeze(real_A_realign.cpu().detach().numpy())
        if(output_image.ndim==2):
            turn=1
        else:
            turn=output_image.shape[0]
        #print(turn)
        if(turn>1):
            for id in range(turn):
                #print('shape of output_image -----> ',output_image.shape)
                aaaa, stack_start_w, stack_end_w, stack_start_h, stack_end_h, stack_start_s= \
                multibatch_test_save(single_coordinate,id,output_image)
                denoise_img[stack_start_s, stack_start_h:stack_end_h, stack_start_w:stack_end_w] \
                = aaaa 

        else:
            aaaa, stack_start_w, stack_end_w, stack_start_h, stack_end_h, stack_start_s = \
            singlebatch_test_save(single_coordinate, output_image)
            # print(stack_start_s)
            denoise_img[stack_start_s, stack_start_h:stack_end_h, stack_start_w:stack_end_w] \
                            = aaaa 
    return denoise_img




import torch
import torch.nn as nn

from deepwonder.SEG.buildingblocks import Encoder, Decoder, FinalConv, DoubleConv, ExtResNetBlock, SingleConv
from deepwonder.SEG.buildingblocks import TS_Encoder, TS_Decoder, DoubleConvTS
from deepwonder.SEG.utils import create_feature_maps


class UNet3D(nn.Module):
    """
    3DUnet model from
    `"3D U-Net: Learning Dense Volumetric Segmentation from Sparse Annotation"
        <https://arxiv.org/pdf/1606.06650.pdf>`.

    Args:
        in_channels (int): number of input channels
        out_channels (int): number of output segmentation masks;
            Note that that the of out_channels might correspond to either
            different semantic classes or to different binary segmentation mask.
            It's up to the user of the class to interpret the out_channels and
            use the proper loss criterion during training (i.e. CrossEntropyLoss (multi-class)
            or BCEWithLogitsLoss (two-class) respectively)
        f_maps (int, tuple): number of feature maps at each level of the encoder; if it's an integer the number
            of feature maps is given by the geometric progression: f_maps ^ k, k=1,2,3,4
        final_sigmoid (bool): if True apply element-wise nn.Sigmoid after the
            final 1x1 convolution, otherwise apply nn.Softmax. MUST be True if nn.BCELoss (two-class) is used
            to train the model. MUST be False if nn.CrossEntropyLoss (multi-class) is used to train the model.
        layer_order (string): determines the order of layers
            in `SingleConv` module. e.g. 'crg' stands for Conv3d+ReLU+GroupNorm3d.
            See `SingleConv` for more info
        init_channel_number (int): number of feature maps in the first conv layer of the encoder; default: 64
        num_groups (int): number of groups for the GroupNorm
    """

    def __init__(self, in_channels, out_channels, final_sigmoid, frame_num, f_maps=64, layer_order='cr', num_groups=8,
                 **kwargs):
        super(UNet3D, self).__init__()

        if isinstance(f_maps, int):
            # use 4 levels in the encoder path as suggested in the paper
            f_maps = create_feature_maps(f_maps, number_of_fmaps=4)

        # create encoder path consisting of Encoder modules. The length of the encoder is equal to `len(f_maps)`
        # uses DoubleConv as a basic_module for the Encoder
        encoders = []
        for i, out_feature_num in enumerate(f_maps):
            if i == 0:
                encoder = Encoder(in_channels, out_feature_num, apply_pooling=False, basic_module=DoubleConv,
                                  conv_layer_order=layer_order, num_groups=num_groups)
            else:
                encoder = Encoder(f_maps[i - 1], out_feature_num, basic_module=DoubleConv,
                                  conv_layer_order=layer_order, num_groups=num_groups)
            encoders.append(encoder)

        self.encoders = nn.ModuleList(encoders)

        # create decoder path consisting of the Decoder modules. The length of the decoder is equal to `len(f_maps) - 1`
        # uses DoubleConv as a basic_module for the Decoder
        decoders = []
        reversed_f_maps = list(reversed(f_maps))
        for i in range(len(reversed_f_maps) - 1):
            in_feature_num = reversed_f_maps[i] + reversed_f_maps[i + 1]
            out_feature_num = reversed_f_maps[i + 1]
            decoder = Decoder(in_feature_num, out_feature_num, basic_module=DoubleConv,
                              conv_layer_order=layer_order, num_groups=num_groups)
            decoders.append(decoder)

        self.decoders = nn.ModuleList(decoders)

        # in the last layer a 1×1 convolution reduces the number of output
        # channels to the number of labels
        self.final_conv = nn.Conv3d(f_maps[0], out_channels, 1)
        # self.final_pooling = nn.Conv3d(out_channels, out_channels,  (128,1,1), padding=0)
        self.final_pooling = nn.MaxPool3d((frame_num,1,1), stride = (frame_num,1,1))
        self.final_activation = nn.Sigmoid()
        if final_sigmoid:
            self.final_activation = nn.Sigmoid()
        else:
            self.final_activation = nn.Softmax(dim=1)

    def forward(self, x):
        # encoder part
        encoders_features = []
        for encoder in self.encoders:
            x = encoder(x)
            # reverse the encoder outputs to be aligned with the decoder
            encoders_features.insert(0, x)

        # remove the last encoder's output from the list
        # !!remember: it's the 1st in the list
        encoders_features = encoders_features[1:]

        # decoder part
        for decoder, encoder_features in zip(self.decoders, encoders_features):
            # pass the output from the corresponding encoder and the output
            # of the previous decoder
            x = decoder(encoder_features, x)

        x = self.final_conv(x)
        # final_pooling = nn.MaxPool3d((x.shape[2],1,1), stride = (x.shape[2],1,1))
        # x = final_pooling(x)
        # print(x.shape)
        x = self.final_pooling(x)
        x = self.final_activation(x)
        # apply final_activation (i.e. Sigmoid or Softmax) only for prediction. During training the network outputs
        # logits and it's up to the user to normalize it before visualising with tensorboard or computing validation metric
        # if not self.training:
            # x = self.final_activation(x)

        return x




class T_conv_net(nn.Module):
    def __init__(self, in_channels, frame_num, tc_f_maps=64, layer_order='cr', num_groups=8,
                 **kwargs):
        super(T_conv_net, self).__init__()

        self.end_frame_num = round(frame_num/8)
        tc_k_size = (3,1,1)
        sc_k_size = (1,3,3)
        padding_size_t = (1,0,0)
        padding_size_s = (0,1,1)
        pool_kernel_size = (2,1,1)
        '''
        self.layer1 = DoubleConv(in_channels, tc_f_maps, encoder=False,
                                kernel_size=tc_k_size, order=layer_order, num_groups=num_groups, padding=padding_size)
        self.layer2 = DoubleConv(tc_f_maps, tc_f_maps, encoder=False,
                        kernel_size=tc_k_size, order=layer_order, num_groups=num_groups, padding=padding_size)
        self.pooling1 = nn.MaxPool3d(kernel_size=pool_kernel_size)
        

        self.layer3 = DoubleConv(tc_f_maps, tc_f_maps, encoder=False,
                        kernel_size=tc_k_size, order=layer_order, num_groups=num_groups, padding=padding_size)
        self.layer4 = DoubleConv(tc_f_maps, tc_f_maps, encoder=False,
                        kernel_size=tc_k_size, order=layer_order, num_groups=num_groups, padding=padding_size)
        self.pooling2 = nn.MaxPool3d(kernel_size=pool_kernel_size)

        self.layer5 = DoubleConv(tc_f_maps, tc_f_maps, encoder=False,
                        kernel_size=tc_k_size, order=layer_order, num_groups=num_groups, padding=padding_size)
        self.layer6 = DoubleConv(tc_f_maps, tc_f_maps, encoder=False,
                        kernel_size=tc_k_size, order=layer_order, num_groups=num_groups, padding=padding_size)
        self.pooling3 = nn.MaxPool3d(kernel_size=pool_kernel_size)

        self.layer7 = DoubleConv(tc_f_maps, tc_f_maps, encoder=False,
                        kernel_size=tc_k_size, order=layer_order, num_groups=num_groups, padding=padding_size)
        self.layer8 = DoubleConv(tc_f_maps, tc_f_maps, encoder=False,
                        kernel_size=tc_k_size, order=layer_order, num_groups=num_groups, padding=padding_size)
        self.final_pooling = nn.MaxPool3d((self.end_frame_num,1,1), stride = (self.end_frame_num,1,1))
        '''
        # self.final_pooling = nn.Conv3d(tc_f_maps, tc_f_maps,  kernel_size=(self.end_frame_num,1,1), stride = (self.end_frame_num,1,1), padding=0)
        self.layer1 = DoubleConvTS(in_channels, tc_f_maps, encoder=False,
                        kernel_size_T=tc_k_size, kernel_size_S=sc_k_size, order=layer_order, 
                        num_groups=num_groups, padding_T=padding_size_t, padding_S=padding_size_s)
        self.layer2 = DoubleConvTS(tc_f_maps, tc_f_maps, encoder=False,
                        kernel_size_T=tc_k_size, kernel_size_S=sc_k_size, order=layer_order, 
                        num_groups=num_groups, padding_T=padding_size_t, padding_S=padding_size_s)
        self.pooling1 = nn.MaxPool3d(kernel_size=pool_kernel_size)
        

        self.layer3 = DoubleConvTS(tc_f_maps, tc_f_maps, encoder=False,
                        kernel_size_T=tc_k_size, kernel_size_S=sc_k_size, order=layer_order, 
                        num_groups=num_groups, padding_T=padding_size_t, padding_S=padding_size_s)
        self.layer4 = DoubleConvTS(tc_f_maps, tc_f_maps, encoder=False,
                        kernel_size_T=tc_k_size, kernel_size_S=sc_k_size, order=layer_order, 
                        num_groups=num_groups, padding_T=padding_size_t, padding_S=padding_size_s)
        self.pooling2 = nn.MaxPool3d(kernel_size=pool_kernel_size)

        self.layer5 = DoubleConvTS(tc_f_maps, tc_f_maps, encoder=False,
                        kernel_size_T=tc_k_size, kernel_size_S=sc_k_size, order=layer_order, 
                        num_groups=num_groups, padding_T=padding_size_t, padding_S=padding_size_s)
        self.layer6 = DoubleConvTS(tc_f_maps, tc_f_maps, encoder=False,
                        kernel_size_T=tc_k_size, kernel_size_S=sc_k_size, order=layer_order, 
                        num_groups=num_groups, padding_T=padding_size_t, padding_S=padding_size_s)
        self.pooling3 = nn.MaxPool3d(kernel_size=pool_kernel_size)

        self.layer7 = DoubleConvTS(tc_f_maps, tc_f_maps, encoder=False,
                        kernel_size_T=tc_k_size, kernel_size_S=sc_k_size, order=layer_order, 
                        num_groups=num_groups, padding_T=padding_size_t, padding_S=padding_size_s)
        self.layer8 = DoubleConvTS(tc_f_maps, tc_f_maps, encoder=False,
                        kernel_size_T=tc_k_size, kernel_size_S=sc_k_size, order=layer_order, 
                        num_groups=num_groups, padding_T=padding_size_t, padding_S=padding_size_s)
        self.final_pooling = nn.MaxPool3d((self.end_frame_num,1,1), stride = (self.end_frame_num,1,1))
    

    def forward(self, x):
        x = self.layer1(x)
        # print('layer1 -----> ',x.shape)
        x = self.layer2(x)
        x = self.pooling1(x)
        # print('layer2 -----> ',x.shape)
        x = self.layer3(x)
        # print('layer3 -----> ',x.shape)
        x = self.layer4(x)
        x = self.pooling2(x)
        # print('layer4 -----> ',x.shape)
        # x = self.final_pooling(x)
        # print('final_pooling -----> ',x.shape)
        x = self.layer5(x)
        x = self.layer6(x)
        x = self.pooling3(x)
        x = self.layer7(x)
        x = self.layer8(x)
        # print('layer8 -----> ',x.shape)
        x = self.final_pooling(x)
        return x



class TS_UNet3D(nn.Module):
    def __init__(self, in_channels, out_channels, final_sigmoid, frame_num, f_maps=64, layer_order='cr', num_groups=8,
                 **kwargs):
        super(TS_UNet3D, self).__init__()

        if isinstance(f_maps, int):
            # use 4 levels in the encoder path as suggested in the paper
            f_maps = create_feature_maps(f_maps, number_of_fmaps=4)

        # create encoder path consisting of Encoder modules. The length of the encoder is equal to `len(f_maps)`
        # uses DoubleConv as a basic_module for the Encoder
        encoders = []
        for i, out_feature_num in enumerate(f_maps):
            if i == 0:
                encoder = TS_Encoder(in_channels, out_feature_num, apply_pooling=False, basic_module=DoubleConv,
                                  conv_layer_order=layer_order, num_groups=num_groups)
            else:
                encoder = TS_Encoder(f_maps[i - 1], out_feature_num, basic_module=DoubleConv,
                                  conv_layer_order=layer_order, num_groups=num_groups)
            encoders.append(encoder)

        self.encoders = nn.ModuleList(encoders)

        # create decoder path consisting of the Decoder modules. The length of the decoder is equal to `len(f_maps) - 1`
        # uses DoubleConv as a basic_module for the Decoder
        decoders = []
        reversed_f_maps = list(reversed(f_maps))
        for i in range(len(reversed_f_maps) - 1):
            in_feature_num = reversed_f_maps[i] + reversed_f_maps[i + 1]
            out_feature_num = reversed_f_maps[i + 1]
            decoder = TS_Decoder(in_feature_num, out_feature_num, basic_module=DoubleConv,
                              conv_layer_order=layer_order, num_groups=num_groups)
            decoders.append(decoder)

        self.decoders = nn.ModuleList(decoders)

        # in the last layer a 1×1 convolution reduces the number of output
        # channels to the number of labels
        self.final_conv = nn.Conv3d(f_maps[0], out_channels, 1)
        # self.final_pooling = nn.Conv3d(out_channels, out_channels,  (128,1,1), padding=0)
        self.final_pooling = nn.MaxPool3d((frame_num,1,1), stride = (frame_num,1,1))
        self.final_activation = nn.Sigmoid()
        if final_sigmoid:
            self.final_activation = nn.Sigmoid()
        else:
            self.final_activation = nn.Softmax(dim=1)

    def forward(self, x):
        # encoder part
        encoders_features = []
        for encoder in self.encoders:
            x = encoder(x)
            # reverse the encoder outputs to be aligned with the decoder
            encoders_features.insert(0, x)

        # remove the last encoder's output from the list
        # !!remember: it's the 1st in the list
        encoders_features = encoders_features[1:]

        # decoder part
        for decoder, encoder_features in zip(self.decoders, encoders_features):
            # pass the output from the corresponding encoder and the output
            # of the previous decoder
            x = decoder(encoder_features, x)

        x = self.final_conv(x)
        # final_pooling = nn.MaxPool3d((x.shape[2],1,1), stride = (x.shape[2],1,1))
        # x = final_pooling(x)
        # print(x.shape)
        x = self.final_pooling(x)
        x = self.final_activation(x)

        return x


import numpy as np
import argparse
import os
import tifffile as tiff
import time
import datetime
from skimage import io
import math
from torch.utils.data import Dataset
import torch


class testset(Dataset):
    def __init__(self,name_list,coordinate_list,noise_img):
        self.name_list = name_list
        self.coordinate_list=coordinate_list
        self.noise_img = noise_img

    def __getitem__(self, index):
        #fn = self.images[index]
        single_coordinate = self.coordinate_list[self.name_list[index]]
        init_h = single_coordinate['init_h']
        end_h = single_coordinate['end_h']
        init_w = single_coordinate['init_w']
        end_w = single_coordinate['end_w']
        init_s = single_coordinate['init_s']
        end_s = single_coordinate['end_s']
        noise_patch = self.noise_img[init_s:end_s, init_h:end_h, init_w:end_w]
        noise_patch=torch.from_numpy(np.expand_dims(noise_patch, 0))
        #target = self.target[index]
        return noise_patch,single_coordinate

    def __len__(self):
        return len(self.name_list)



def singlebatch_test_save(single_coordinate,output_image):
    stack_start_w = int(single_coordinate['stack_start_w'])
    stack_end_w = int(single_coordinate['stack_end_w'])
    patch_start_w = int(single_coordinate['patch_start_w'])
    patch_end_w = int(single_coordinate['patch_end_w'])

    stack_start_h = int(single_coordinate['stack_start_h'])
    stack_end_h = int(single_coordinate['stack_end_h'])
    patch_start_h = int(single_coordinate['patch_start_h'])
    patch_end_h = int(single_coordinate['patch_end_h'])

    stack_start_s = int(single_coordinate['stack_start_s'])

    aaaa = output_image[patch_start_h:patch_end_h, patch_start_w:patch_end_w]
    return aaaa,stack_start_w,stack_end_w,stack_start_h,stack_end_h,stack_start_s


def multibatch_test_save(single_coordinate,id,output_image):
    stack_start_w_id = single_coordinate['stack_start_w'].numpy()
    stack_start_w = int(stack_start_w_id[id])
    stack_end_w_id = single_coordinate['stack_end_w'].numpy()
    stack_end_w=int(stack_end_w_id[id])
    patch_start_w_id = single_coordinate['patch_start_w'].numpy()
    patch_start_w=int(patch_start_w_id[id])
    patch_end_w_id = single_coordinate['patch_end_w'].numpy()
    patch_end_w=int(patch_end_w_id[id])

    stack_start_h_id = single_coordinate['stack_start_h'].numpy()
    stack_start_h = int(stack_start_h_id[id])
    stack_end_h_id = single_coordinate['stack_end_h'].numpy()
    stack_end_h = int(stack_end_h_id[id])
    patch_start_h_id = single_coordinate['patch_start_h'].numpy()
    patch_start_h = int(patch_start_h_id[id])
    patch_end_h_id = single_coordinate['patch_end_h'].numpy()
    patch_end_h = int(patch_end_h_id[id])

    stack_start_s_id = single_coordinate['stack_start_s'].numpy()
    stack_start_s = int(stack_start_s_id[id])

    output_image_id=output_image[id]
    aaaa = output_image_id[patch_start_h:patch_end_h, patch_start_w:patch_end_w]

    return aaaa,stack_start_w,stack_end_w,stack_start_h,stack_end_h,stack_start_s



def test_preprocess_lessMemoryNoTail_SubImgSEG(args, sub_img):
    img_h = args['img_h']
    img_w = args['img_w']
    img_s2 = args['img_s']
    gap_h = args['gap_h']
    gap_w = args['gap_w']
    gap_s2 = args['gap_s']
    cut_w = (img_w - gap_w)/2
    cut_h = (img_h - gap_h)/2
    cut_s = (img_s2 - gap_s2)/2

    noise_im = sub_img
    noise_im = (noise_im).astype(np.float32)/args['normalize_factor']

    noise_im_ave_single = np.mean(noise_im, axis=0)
    noise_im_ave = np.zeros(noise_im.shape)
    for i in range(0, noise_im.shape[0]):
        noise_im_ave[i,:,:] = noise_im_ave_single
    noise_im = noise_im-noise_im_ave

    whole_w = noise_im.shape[2]
    whole_h = noise_im.shape[1]
    whole_s = noise_im.shape[0]

    num_w = math.ceil((whole_w-img_w+gap_w)/gap_w)
    num_h = math.ceil((whole_h-img_h+gap_h)/gap_h)
    num_s = math.ceil((whole_s-img_s2+gap_s2)/gap_s2)
    # print('int((whole_h-img_h+gap_h)/gap_h) -----> ',int((whole_h-img_h+gap_h)/gap_h))
    # print('int((whole_w-img_w+gap_w)/gap_w) -----> ',int((whole_w-img_w+gap_w)/gap_w))
    # print('int((whole_s-img_s2+gap_s2)/gap_s2) -----> ',int((whole_s-img_s2+gap_s2)/gap_s2))
    name_list = []
    coordinate_list = {}
    for x in range(0,num_h):
        for y in range(0,num_w):
            for z in range(0,num_s):
                single_coordinate={'init_h':0, 'end_h':0, 'init_w':0, 'end_w':0, 'init_s':0, 'end_s':0}
                if x != (num_h-1):
                    init_h = gap_h*x
                    end_h = gap_h*x + img_h
                elif x == (num_h-1):
                    init_h = whole_h - img_h
                    end_h = whole_h

                if y != (num_w-1):
                    init_w = gap_w*y
                    end_w = gap_w*y + img_w
                elif y == (num_w-1):
                    init_w = whole_w - img_w
                    end_w = whole_w

                if z != (num_s-1):
                    init_s = gap_s2*z
                    end_s = gap_s2*z + img_s2
                elif z == (num_s-1):
                    init_s = whole_s - img_s2
                    end_s = whole_s
                single_coordinate['init_h'] = init_h
                single_coordinate['end_h'] = end_h
                single_coordinate['init_w'] = init_w
                single_coordinate['end_w'] = end_w
                single_coordinate['init_s'] = init_s
                single_coordinate['end_s'] = end_s

                if (num_w-1)==0:
                    single_coordinate['stack_start_w'] = 0
                    single_coordinate['stack_end_w'] = whole_w
                    single_coordinate['patch_start_w'] = 0
                    single_coordinate['patch_end_w'] = img_w
                elif (num_w-1)>0:
                    if y == 0:
                        single_coordinate['stack_start_w'] = y*gap_w
                        single_coordinate['stack_end_w'] = y*gap_w+img_w-cut_w
                        single_coordinate['patch_start_w'] = 0
                        single_coordinate['patch_end_w'] = img_w-cut_w
                    elif y == num_w-1:
                        single_coordinate['stack_start_w'] = whole_w-img_w+cut_w
                        single_coordinate['stack_end_w'] = whole_w
                        single_coordinate['patch_start_w'] = cut_w
                        single_coordinate['patch_end_w'] = img_w
                    else:
                        single_coordinate['stack_start_w'] = y*gap_w+cut_w
                        single_coordinate['stack_end_w'] = y*gap_w+img_w-cut_w
                        single_coordinate['patch_start_w'] = cut_w
                        single_coordinate['patch_end_w'] = img_w-cut_w

                if (num_h-1)==0:
                    single_coordinate['stack_start_h'] = 0
                    single_coordinate['stack_end_h'] = whole_h
                    single_coordinate['patch_start_h'] = 0
                    single_coordinate['patch_end_h'] = img_h
                elif (num_h-1)>0:
                    if x == 0:
                        single_coordinate['stack_start_h'] = x*gap_h
                        single_coordinate['stack_end_h'] = x*gap_h+img_h-cut_h
                        single_coordinate['patch_start_h'] = 0
                        single_coordinate['patch_end_h'] = img_h-cut_h
                    elif x == num_h-1:
                        single_coordinate['stack_start_h'] = whole_h-img_h+cut_h
                        single_coordinate['stack_end_h'] = whole_h
                        single_coordinate['patch_start_h'] = cut_h
                        single_coordinate['patch_end_h'] = img_h
                    else:
                        single_coordinate['stack_start_h'] = x*gap_h+cut_h
                        single_coordinate['stack_end_h'] = x*gap_h+img_h-cut_h
                        single_coordinate['patch_start_h'] = cut_h
                        single_coordinate['patch_end_h'] = img_h-cut_h

                single_coordinate['stack_start_s'] = z


                # noise_patch1 = noise_im[init_s:end_s,init_h:end_h,init_w:end_w]
                patch_name = 'x'+str(x)+'_y'+str(y)+'_z'+str(z)
                # train_raw.append(noise_patch1.transpose(1,2,0))
                name_list.append(patch_name)
                # print(' single_coordinate -----> ',single_coordinate)
                coordinate_list[patch_name] = single_coordinate

    return  name_list, noise_im, coordinate_list



def get_gap_s(args, img, stack_num):
    whole_w = img.shape[2]
    whole_h = img.shape[1]
    whole_s = img.shape[0]
    print('whole_w -----> ',whole_w)
    print('whole_h -----> ',whole_h)
    print('whole_s -----> ',whole_s)
    w_num = math.floor((whole_w-args.img_w)/args.gap_w)+1
    h_num = math.floor((whole_h-args.img_h)/args.gap_h)+1
    s_num = math.ceil(args.train_datasets_size/w_num/h_num/stack_num)
    print('w_num -----> ',w_num)
    print('h_num -----> ',h_num)
    print('s_num -----> ',s_num)
    gap_s = math.floor((whole_s-args.img_s*2)/(s_num-1))
    print('gap_s -----> ',gap_s)
    return gap_s




from deepwonder.RMBG.model_3DUnet import UNet3D
import torch.nn as nn

class Network_3D_Unet(nn.Module):
    def __init__(self, UNet_type = '3DUNet', in_channels=1, out_channels=1, f_maps=64, final_sigmoid = True):
        super(Network_3D_Unet, self).__init__()

        self.in_channels = in_channels
        self.out_channels = out_channels
        self.final_sigmoid = final_sigmoid

        if UNet_type == '3DUNet':
            self.Generator = UNet3D( in_channels = in_channels,
                                     out_channels = out_channels,
                                     f_maps = f_maps, 
                                     final_sigmoid = final_sigmoid)

    def forward(self, x):
        fake_x = self.Generator(x)
        return fake_x



import cv2
import numpy as np
import scipy.io as scio

def neighbor_label_list(i, j, label_matrix, binary_img, neighbor_hoods):
    nb_label_list=[]

    if neighbor_hoods == 'NEIGHBOR_HOODS_4':
        if i > 0:
            label_up = label_matrix[i-1, j].copy()
            pixel_up = binary_img[i-1, j].copy()
            if pixel_up!=0:
                if label_up not in nb_label_list:
                    nb_label_list.append(label_up)
        if i < label_matrix.shape[0]-1:
            label_down = label_matrix[i+1, j].copy()
            pixel_down = binary_img[i+1, j].copy()
            if pixel_down!=0:
                if label_down not in nb_label_list:
                    nb_label_list.append(label_down)
        if j > 0:
            label_left = label_matrix[i, j-1].copy()
            pixel_left = binary_img[i, j-1].copy()
            if pixel_left!=0:
                if label_left not in nb_label_list:
                    nb_label_list.append(label_left)
        if j < label_matrix.shape[1]-1:
            label_right = label_matrix[i, j+1].copy()
            pixel_right = binary_img[i, j+1].copy()
            if pixel_right!=0:
                if label_right not in nb_label_list:
                    nb_label_list.append(label_right)

    if neighbor_hoods == 'NEIGHBOR_HOODS_8':
        if i > 0:
            label_up = label_matrix[i-1, j].copy()
            pixel_up = binary_img[i-1, j].copy()
            if pixel_up!=0:
                if label_up not in nb_label_list:
                    nb_label_list.append(label_up)
        if i < label_matrix.shape[0]-1:
            label_down = label_matrix[i+1, j].copy()
            pixel_down = binary_img[i+1, j].copy()
            if pixel_down!=0:
                if label_down not in nb_label_list:
                    nb_label_list.append(label_down)
        if j > 0:
            label_left = label_matrix[i, j-1].copy()
            pixel_left = binary_img[i, j-1].copy()
            if pixel_left!=0:
                if label_left not in nb_label_list:
                    nb_label_list.append(label_left)
        if j < label_matrix.shape[1]-1:
            label_right = label_matrix[i, j+1].copy()
            pixel_right = binary_img[i, j+1].copy()
            if pixel_right!=0:
                if label_right not in nb_label_list:
                    nb_label_list.append(label_right)

        if (i > 0) and (j > 0):
            label_up_left = label_matrix[i-1, j-1].copy()
            pixel_up_left = binary_img[i-1, j-1].copy()
            if pixel_up_left!=0:
                if label_up_left not in nb_label_list:
                    nb_label_list.append(label_up_left)
        if (i > 0) and (j < label_matrix.shape[1]-1):
            label_up_right = label_matrix[i-1, j+1].copy()
            pixel_up_right = binary_img[i-1, j+1].copy()
            if pixel_up_right!=0:
                if label_up_right not in nb_label_list:
                    nb_label_list.append(label_up_right)
        if (i < label_matrix.shape[0]-1) and (j > 0):
            label_down_left = label_matrix[i+1,  j-1].copy()
            pixel_down_left = binary_img[i+1,  j-1].copy()
            if pixel_down_left!=0:
                if label_down_left not in nb_label_list:
                    nb_label_list.append(label_down_left)
        if (i < label_matrix.shape[0]-1) and (j < label_matrix.shape[1]-1):
            label_down_right = label_matrix[i+1, j+1].copy()
            pixel_down_right = binary_img[i+1, j+1].copy()
            if pixel_down_right!=0:
                if label_down_right not in nb_label_list:
                    nb_label_list.append(label_down_right)
    return nb_label_list


def New_Two_Pass(binary_img: np.array, neighbor_hoods):
    if neighbor_hoods == 'NEIGHBOR_HOODS_4':
        pass
    elif neighbor_hoods == 'NEIGHBOR_HOODS_8':
        pass
    else:
        raise ValueError

    rows, cols = binary_img.shape
    label_matrix = np.zeros((rows, cols))
    label_idx = 1
    for i in range(0,rows):
        for j in range(0,cols):
            pixel_center = binary_img[i,j]
            if pixel_center !=0:
                
                nb_label_list = neighbor_label_list(i, j, label_matrix, binary_img, neighbor_hoods)
                # print('i --- ',i,' j --- ',j,' pixel_center --- ',pixel_center,' nb_label_list --- ',nb_label_list, \
                # ' label_matrix04 --- ',label_matrix[4,0],np.min(label_matrix),np.max(label_matrix))
                if (len(nb_label_list)==1) and (min(nb_label_list)==0):
                    label_matrix[i,j]=label_idx
                    label_idx = label_idx+1
                    # print('i --- ',i,' j --- ',j,' label_idex --- ',label_idx)
                if (len(nb_label_list)==1) and (min(nb_label_list)!=0):
                    label_matrix[i,j]=min(nb_label_list)
                if (len(nb_label_list)>1):
                    if 0 in nb_label_list:
                        if (len(nb_label_list)==2):
                            nb_label_list.remove(0)
                            if min(nb_label_list)!=0:
                                label_matrix[i,j]=min(nb_label_list)
                        if (len(nb_label_list)>2):
                            # if 0 in nb_label_list:
                            nb_label_list.remove(0)
                            if min(nb_label_list)!=0:
                                label_matrix[i,j]=min(nb_label_list)
                                min_nb_label_list=min(nb_label_list)
                                nb_label_list.remove(min_nb_label_list)
                                # print('i --- ',i,' j --- ',j,' pixel_center --- ',pixel_center,' remove nb_label_list --- ',nb_label_list)
                                for idx_nb_label_list in range(0, len(nb_label_list)):
                                    nb_value = nb_label_list[idx_nb_label_list]
                                    # print('nb_value --- ',nb_value)
                                    # print('np.argwhere(label_matrix == nb_value) --- ',np.argwhere(label_matrix == nb_value))
                                    # print('np.argwhere(label_matrix == nb_value) --- ',np.argwhere(label_matrix == 3))
                                    label_matrix[label_matrix == nb_value] = min_nb_label_list
                                    # print('np.argwhere(label_matrix == nb_value) --- ',np.argwhere(label_matrix == 3))
                    if 0 not in nb_label_list:
                        label_matrix[i,j]=min(nb_label_list)
                        min_nb_label_list=min(nb_label_list)
                        nb_label_list.remove(min_nb_label_list)
                        for idx_nb_label_list in range(0, len(nb_label_list)):
                            nb_value = nb_label_list[idx_nb_label_list]
                            label_matrix[label_matrix == nb_value] = min_nb_label_list
    return label_matrix


from deepwonder.SEG.model_3DUnet import UNet3D, T_conv_net, TS_UNet3D
import torch.nn as nn
import torch.nn.functional as F
import torch
from deepwonder.SEG.model_Unet import Unet, Unet3, Unet4



class SEG_Network_3D_Unet(nn.Module):
    def __init__(self, UNet_type = '3DUNet', in_channels=1, out_channels=1, frame_num=64, final_sigmoid = True,f_maps=64):
        super(SEG_Network_3D_Unet, self).__init__()

        self.in_channels = in_channels
        self.out_channels = out_channels
        self.final_sigmoid = final_sigmoid

        if UNet_type == '3DUNet':
            self.Generator = UNet3D( in_channels = in_channels,
                                     out_channels = out_channels, 
                                     frame_num = frame_num,
                                     final_sigmoid = final_sigmoid,
                                     f_maps = f_maps)
        if UNet_type == 'TS_UNet3D':
            self.Generator = TS_UNet3D( in_channels = in_channels,
                                     out_channels = out_channels, 
                                     frame_num = frame_num,
                                     final_sigmoid = final_sigmoid,
                                     f_maps = f_maps)
    def forward(self, x):
        fake_x = self.Generator(x)
        return fake_x


class Network_TSnet(nn.Module):
    def __init__(self,  in_channels=1, out_channels=1, frame_num=64, f_maps=64):
        super(Network_TSnet, self).__init__()

        self.in_channels = in_channels
        self.out_channels = out_channels

        self.T_conv_net = T_conv_net( in_channels = in_channels, 
                                    frame_num = frame_num,
                                    tc_f_maps = f_maps)
        self.Unet = Unet3( in_ch = f_maps, 
                                out_ch = out_channels,
                                    f_num = f_maps)

        # T_conv_net = T_conv_net(in_channels = 1, frame_num = opt.frame_num, tc_f_maps=opt.tc_f_num)
        # Unet_1 = Unet3(in_ch = opt.tc_f_num, out_ch = 1, f_num = opt.unet_f_num)

    def forward(self, x):
        # print('Initialize TS_net --->')
        pred_patch = self.T_conv_net(x)
        input_u = torch.squeeze(pred_patch, dim=2)
        output_u = self.Unet(input_u)
        # fake_x = self.Generator(x)
        return output_u



import tifffile as tiff
import cv2
import numpy as np
from skimage import io

import time
import datetime
import scipy.io as scio
import multiprocessing as mp
import math
import os

from deepwonder.MN.utils import Split_Neuron, listAddtrace
from deepwonder.MN.utils import Neuron_List_Initial, list2mask, listAddtrace, centroid_distance, list_union, initial_mask_list
from deepwonder.MN.utils import list2contours, Joint_Mask_List_Simple, Joint_Mask_List_Mul
from deepwonder.MN.utils import listAddcontours_Laplacian_pytorch, list2contours
from deepwonder.MN.utils import group_mask, calculate_trace, Mining_rest_neuron, clear_neuron

def merge_neuron_SEG_mul(mask_stack, 
                    raw_image, 
                    quit_round_rate = 0.5, 
                    good_round_rate = 0.8,
                    good_round_size_rate = 0.4, 
                    corr_mark=0.9, 
                    max_value=1000,
                    if_nmf = True):
    '''
    save_folder = datetime.datetime.now().strftime("%Y%m%d-%H%M")
    if not os.path.exists(save_folder): 
        os.mkdir(save_folder)
    '''

    threshold = max_value*0.3
    mask_stack[mask_stack>threshold] = max_value
    mask_stack[mask_stack<threshold] = 0

    w_good_neuron_list = []
    w_bad_neuron_list = []
    g_f_contours = np.zeros(mask_stack.shape)
    b_f_contours = np.zeros(mask_stack.shape)

    prev_time = time.time()
    time_start = time.time()

    if_mul=1
    if if_mul:
        num_cores = int(mp.cpu_count())
        pool = mp.Pool(num_cores)
        mask_dict = {}
        for i in range(0, mask_stack.shape[0]):
            mask_dict[str(i)] = mask_stack[i,:,:].squeeze()
        results = [pool.apply_async(initial_mask_list, args=(mask, 0.5, 0.85, 60)) for name, mask in mask_dict.items()]
        for p in results:
            good_neuron_list, bad_neuron_list = p.get()
            w_good_neuron_list = good_neuron_list + w_good_neuron_list
            w_bad_neuron_list = w_bad_neuron_list + bad_neuron_list
            # print('good_neuron_list ---> ',len(good_neuron_list),' bad_neuron_list ---> ',len(bad_neuron_list))
        time_end = time.time()
        time_cost = time_end - time_start
        print('[Neuron Segmentation Mask Initialize Time Cost: %.0d s] \n'%(time_cost), end=' ')
        pool.close()
        pool.join()
    if not if_mul:
        for i in range(0, mask_stack.shape[0]):
            now_mask = mask_stack[i,:,:].squeeze()
            good_neuron_list, bad_neuron_list = initial_mask_list(now_mask, 
                                                            quit_round_rate = 0.5, 
                                                            good_round_rate = 0.85, 
                                                            smallest_neuron_area = 60)

            w_good_neuron_list = good_neuron_list + w_good_neuron_list
            w_bad_neuron_list = w_bad_neuron_list + bad_neuron_list
            ################################################################################################################
            # Determine approximate time left
            batches_done = i+1
            batches_left = 1 * mask_stack.shape[0] - batches_done
            time_left_seconds = int(batches_left * (time.time() - prev_time))
            time_left = datetime.timedelta(seconds=time_left_seconds)
            prev_time = time.time()
            ################################################################################################################
            if i % 1 == 0:
                time_end = time.time()
                time_cost = time_end - time_start  # datetime.timedelta(seconds= (time_end - time_start))
                print(
                    '\r\033[1;33m[MERGE]\033[0m [Patch %d/%d] [Time Cost: %.0d s] [ETA: %s s]     '
                    % ( 
                    i ,
                    mask_stack.shape[0],
                    time_cost,
                    time_left_seconds
                    ), end=' ')
            if (i + 1) %  mask_stack.shape[0] == 0:
                print('\n', end=' ')
            ################################################################################################################

    w_good_neuron_list = listAddtrace(w_good_neuron_list, raw_image)

    if_time = True
    if if_time:
        time_cost = time.time() - time_start
        print('[1 Time Cost: %.0d s] \n'%(time_cost), end=' ')
    
    if_split = True
    if not if_split:
        w_good_neuron_list = Joint_Mask_List_Simple(w_good_neuron_list, [], corr_mark=0.9, area_mark=0.9, active_rate=0, if_coor=True, if_area=True, if_merge=False)
    if if_split:
        w_good_neuron_list = Joint_Mask_List_Mul(w_good_neuron_list, corr_mark=0.9, area_mark=0.9, active_rate=0, if_coor=True, if_area=True, if_merge=False)

    if if_time:
        time_cost = time.time() - time_start
        print('[2 Time Cost: %.0d s] \n'%(time_cost), end=' ')
    # w_good_neuron_list = listAddtrace(w_good_neuron_list, raw_image)
    # print('w_good_neuron_list ---> ',len(w_good_neuron_list))
    w_good_neuron_list = clear_neuron(w_good_neuron_list, [], area_mark=0.2, area_size=100)

    if if_time:
        time_cost = time.time() - time_start
        print('[3 Time Cost: %.0d s] \n'%(time_cost), end=' ')

    '''
    if not if_split:
        w_good_neuron_list = Joint_Mask_List_Simple(w_good_neuron_list, [], corr_mark=0.7, area_mark=0.6, active_rate=0, if_coor=False, if_area=True, if_merge=False)
    if if_split:
        w_good_neuron_list = Joint_Mask_List_Mul(w_good_neuron_list, corr_mark=0.7, area_mark=0.6, active_rate=0, if_coor=False, if_area=True, if_merge=False)
    '''
    # print('w_good_neuron_list ---> ',len(w_good_neuron_list))
    # w_good_neuron_list = Joint_Mask_List_Simple(w_good_neuron_list, [], corr_mark=0.7, area_mark=0.6, active_rate=0, if_coor=False, if_area=True, if_merge=False)
    
    
    if if_time:
        time_cost = time.time() - time_start
        print('[4 Time Cost: %.0d s] \n'%(time_cost), end=' ')
    # ,  save_folder=save_folder
    w_bad_neuron_list = listAddtrace(w_bad_neuron_list, raw_image)
    if not if_split:
        w_bad_neuron_list = Joint_Mask_List_Simple(w_bad_neuron_list, [], corr_mark=0.9, area_mark=0.9, active_rate=0, if_coor=True, if_area=True, if_merge=True)
    if if_split:
        w_bad_neuron_list = Joint_Mask_List_Mul(w_bad_neuron_list, corr_mark=0.9, area_mark=0.9, active_rate=0, if_coor=True, if_area=True, if_merge=True)

    if if_time:
        time_cost = time.time() - time_start
        print('[5 Time Cost: %.0d s] \n'%(time_cost), end=' ')

    add_neuron_list = Mining_rest_neuron(w_good_neuron_list, w_bad_neuron_list, raw_image, quit_round_rate = 0.5, smallest_neuron_area = 200)
    add_neuron_list = listAddtrace(add_neuron_list, raw_image)
    if not if_split:
        add_neuron_list = Joint_Mask_List_Simple(add_neuron_list, [], corr_mark=0.9, area_mark=0.9, active_rate=0, if_coor=True, if_area=True, if_merge=True)
    if if_split:
        add_neuron_list = Joint_Mask_List_Mul(add_neuron_list, corr_mark=0.9, area_mark=0.9, active_rate=0, if_coor=True, if_area=True, if_merge=True)
    
    if if_time:
        time_cost = time.time() - time_start
        print('[6 Time Cost: %.0d s] \n'%(time_cost), end=' ')

    w_g_neuron_list = w_good_neuron_list+add_neuron_list
    # corr_mark, area_mark=0.9, active_rate=0.02, if_coor=True, if_area=True, if_merge=True)

    # w_g_mask_list = listAddcontours_Laplacian_pytorch(w_g_neuron_list, mask_stack.shape[1], mask_stack.shape[2])
    # w_g_f_contours, w_g_w_contours = list2contours(w_g_mask_list, mask_stack.shape[1], mask_stack.shape[2])
    # io.imsave(save_folder+'//'+'good.tif', w_g_f_contours)
    # io.imsave(save_folder+'//'+'good_w.tif', w_g_w_contours)


    # print('w_good_neuron_list -----> ',len(w_good_neuron_list))
    if if_nmf:
        group_neuron_list, arranged_index, cc_mask_j = group_mask(w_g_neuron_list, raw_image)
        '''
        group_neuron_list1 = listAddcontours_Laplacian_pytorch(group_neuron_list, mask_stack.shape[1], mask_stack.shape[2])
        g_n_f_contours, g_n_w_contours = list2contours(group_neuron_list1, mask_stack.shape[1], mask_stack.shape[2])
        io.imsave(save_folder+'//'+'group.tif', g_n_f_contours)
        io.imsave(save_folder+'//'+'cc_mask_j.tif', cc_mask_j)
        '''
        w_g_neuron_list = calculate_trace(raw_image, group_neuron_list, w_g_neuron_list, arranged_index)

    # print('w_good_neuron_list -----> ',len(w_good_neuron_list))
    # return final_mask_list_ct
    '''
    mat_save_name = save_folder+'//'+'neuron_list.mat'
    scio.savemat(mat_save_name, {'w_good_neuron_list':w_good_neuron_list})
    '''
    w_g_neuron_list = Joint_Mask_List_Simple(w_g_neuron_list, [], corr_mark=0.7, area_mark=0.6, active_rate=0, if_coor=False, if_area=True, if_merge=False)

    return w_g_neuron_list



def merge_neuron_SEG(mask_stack, 
                    raw_image, 
                    quit_round_rate = 0.5, 
                    good_round_rate = 0.8,
                    good_round_size_rate = 0.4, 
                    corr_mark=0.9, 
                    max_value=1000,
                    if_nmf = True):

    save_folder = datetime.datetime.now().strftime("%Y%m%d-%H%M")
    if not os.path.exists(save_folder): 
        os.mkdir(save_folder)


    threshold = max_value*0.3
    mask_stack[mask_stack>threshold] = max_value
    mask_stack[mask_stack<threshold] = 0

    w_good_neuron_list = []
    w_bad_neuron_list = []
    g_f_contours = np.zeros(mask_stack.shape)
    b_f_contours = np.zeros(mask_stack.shape)

    prev_time = time.time()
    time_start = time.time()
    for i in range(0, mask_stack.shape[0]):
        now_mask = mask_stack[i,:,:].squeeze()
        good_neuron_list, bad_neuron_list = initial_mask_list(now_mask, 
                                                        quit_round_rate = 0.5, 
                                                        good_round_rate = 0.85, 
                                                        smallest_neuron_area = 60)

        # good_mask_list = listAddcontours_Laplacian_pytorch(good_neuron_list, mask_stack.shape[1], mask_stack.shape[2])
        # good_final_contours, good_whole_contours = list2contours(good_mask_list, mask_stack.shape[1], mask_stack.shape[2])
        # g_f_contours[i,:,:] = good_final_contours
        # io.imsave(save_folder+'//'+str(i)+'_good.tif', good_final_contours)

        # bad_mask_list = listAddcontours_Laplacian_pytorch(bad_neuron_list, mask_stack.shape[1], mask_stack.shape[2])
        # bad_final_contours, bad_whole_contours = list2contours(bad_mask_list, mask_stack.shape[1], mask_stack.shape[2])
        # b_f_contours[i,:,:] = bad_final_contours
        # io.imsave(save_folder+'//'+str(i)+'_bad.tif', bad_final_contours)

        w_good_neuron_list = good_neuron_list + w_good_neuron_list
        w_bad_neuron_list = w_bad_neuron_list + bad_neuron_list
        ################################################################################################################
        # Determine approximate time left
        batches_done = i+1
        batches_left = 1 * mask_stack.shape[0] - batches_done
        time_left_seconds = int(batches_left * (time.time() - prev_time))
        time_left = datetime.timedelta(seconds=time_left_seconds)
        prev_time = time.time()
        ################################################################################################################
        if i % 1 == 0:
            time_end = time.time()
            time_cost = time_end - time_start  # datetime.timedelta(seconds= (time_end - time_start))
            print(
                '\r\033[1;33m[MERGE]\033[0m [Patch %d/%d] [Time Cost: %.0d s] [ETA: %s s]     '
                % ( 
                i ,
                mask_stack.shape[0],
                time_cost,
                time_left_seconds
                ), end=' ')
        if (i + 1) %  mask_stack.shape[0] == 0:
            print('\n', end=' ')
        ################################################################################################################

    # io.imsave(save_folder+'//'+'g_f_contours.tif', g_f_contours)
    # io.imsave(save_folder+'//'+'b_f_contours.tif', b_f_contours)

    w_good_neuron_list = listAddtrace(w_good_neuron_list, raw_image)
    w_good_neuron_list = Joint_Mask_List_Simple(w_good_neuron_list, [], corr_mark=0.9, area_mark=0.9, active_rate=0, if_coor=True, if_area=True, if_merge=False)
    # w_good_neuron_list = listAddtrace(w_good_neuron_list, raw_image)
    print('w_good_neuron_list ---> ',len(w_good_neuron_list))
    w_good_neuron_list = clear_neuron(w_good_neuron_list, [], area_mark=0.2, area_size=100)
    w_good_neuron_list = Joint_Mask_List_Simple(w_good_neuron_list, [], corr_mark=0.7, area_mark=0.7, active_rate=0, if_coor=False, if_area=True, if_merge=False)
    print('w_good_neuron_list ---> ',len(w_good_neuron_list))
    # w_good_neuron_list = Joint_Mask_List_Simple(w_good_neuron_list, [], corr_mark=0.7, area_mark=0.6, active_rate=0, if_coor=False, if_area=True, if_merge=False)

    # ,  save_folder=save_folder
    w_bad_neuron_list = listAddtrace(w_bad_neuron_list, raw_image)
    w_bad_neuron_list = Joint_Mask_List_Simple(w_bad_neuron_list, [], corr_mark=0.9, area_mark=0.9, active_rate=0, if_coor=True, if_area=True, if_merge=True)
    add_neuron_list = Mining_rest_neuron(w_good_neuron_list, w_bad_neuron_list, raw_image, quit_round_rate = 0.5, smallest_neuron_area = 200)
    add_neuron_list = listAddtrace(add_neuron_list, raw_image)
    add_neuron_list = Joint_Mask_List_Simple(add_neuron_list, [], corr_mark=0.9, area_mark=0.9, active_rate=0, if_coor=True, if_area=True, if_merge=True)
    
    w_g_neuron_list = w_good_neuron_list+add_neuron_list
    # corr_mark, area_mark=0.9, active_rate=0.02, if_coor=True, if_area=True, if_merge=True)

    w_g_mask_list = listAddcontours_Laplacian_pytorch(w_g_neuron_list, mask_stack.shape[1], mask_stack.shape[2])
    w_g_f_contours, w_g_w_contours = list2contours(w_g_mask_list, mask_stack.shape[1], mask_stack.shape[2])
    io.imsave(save_folder+'//'+'good.tif', w_g_f_contours)
    io.imsave(save_folder+'//'+'good_w.tif', w_g_w_contours)


    # print('w_good_neuron_list -----> ',len(w_good_neuron_list))
    if if_nmf:
        group_neuron_list, arranged_index, cc_mask_j = group_mask(w_good_neuron_list, raw_image)
        '''
        group_neuron_list1 = listAddcontours_Laplacian_pytorch(group_neuron_list, mask_stack.shape[1], mask_stack.shape[2])
        g_n_f_contours, g_n_w_contours = list2contours(group_neuron_list1, mask_stack.shape[1], mask_stack.shape[2])
        io.imsave(save_folder+'//'+'group.tif', g_n_f_contours)
        io.imsave(save_folder+'//'+'cc_mask_j.tif', cc_mask_j)
        '''
        w_good_neuron_list = calculate_trace(raw_image, group_neuron_list, w_good_neuron_list, arranged_index)

    # print('w_good_neuron_list -----> ',len(w_good_neuron_list))
    # return final_mask_list_ct
    '''
    mat_save_name = save_folder+'//'+'neuron_list.mat'
    scio.savemat(mat_save_name, {'w_good_neuron_list':w_good_neuron_list})
    '''
    return w_good_neuron_list







import torch
########################################################################################################################
def create_feature_maps(init_channel_number, number_of_fmaps):
    return [init_channel_number * 2 ** k for k in range(number_of_fmaps)]

def FFDrealign4(input):
    # batch channel time height width
    realign_input = torch.cuda.FloatTensor(input.shape[0], input.shape[1]*4, input.shape[2], int(input.shape[3]/2), int(input.shape[4]/2))
    # print('realign_input -----> ',realign_input.shape)
    # print('input -----> ',input.shape)
    # print('realign_input[:,0,:,:,:] -----> ',realign_input[:,0,:,:,:].shape)
    # print('input[:,:,:,0::2, 0::2] -----> ',input[:,:,:,0::2, 0::2].shape)
    realign_input[:,0,:,:,:] = torch.squeeze(input[:,:,:,0::2, 0::2])
    realign_input[:,1,:,:,:] = torch.squeeze(input[:,:,:,0::2, 1::2])
    realign_input[:,2,:,:,:] = torch.squeeze(input[:,:,:,1::2, 0::2])
    realign_input[:,3,:,:,:] = torch.squeeze(input[:,:,:,1::2, 1::2])
    return realign_input

def inv_FFDrealign4(input):
    # batch channel time height width
    realign_input = torch.cuda.FloatTensor(input.shape[0], int(input.shape[1]/4), input.shape[2], int(input.shape[3]*2), int(input.shape[4]*2))

    # print('realign_input[:,:,:,0::2, 0::2] -----> ',realign_input[:,:,:,0::2, 0::2].shape)
    # print('input[:,0,:,:,:] -----> ',input[:,0,:,:,:].shape)
    realign_input[:,:,:,0::2, 0::2] = torch.unsqueeze(input[:,0,:,:,:], 1)
    realign_input[:,:,:,0::2, 1::2] = torch.unsqueeze(input[:,1,:,:,:], 1)
    realign_input[:,:,:,1::2, 0::2] = torch.unsqueeze(input[:,2,:,:,:], 1)
    realign_input[:,:,:,1::2, 1::2] = torch.unsqueeze(input[:,3,:,:,:], 1)
    return realign_input


def FFDrealign8(input):
    # batch channel time height width
    realign_input = torch.cuda.FloatTensor(input.shape[0], input.shape[1]*8, int(input.shape[2]/2), int(input.shape[3]/2), int(input.shape[4]/2))
    # print('realign_input -----> ',realign_input.shape)
    # print('input -----> ',input.shape)
    realign_input[:,0,:,:,:] = input[:,:,0::2,0::2, 0::2]
    realign_input[:,1,:,:,:] = input[:,:,0::2,0::2, 1::2]
    realign_input[:,2,:,:,:] = input[:,:,0::2,1::2, 0::2]
    realign_input[:,3,:,:,:] = input[:,:,0::2,1::2, 1::2]

    realign_input[:,4,:,:,:] = input[:,:,1::2,0::2, 0::2]
    realign_input[:,5,:,:,:] = input[:,:,1::2,0::2, 1::2]
    realign_input[:,6,:,:,:] = input[:,:,1::2,1::2, 0::2]
    realign_input[:,7,:,:,:] = input[:,:,1::2,1::2, 1::2]
    return realign_input

def inv_FFDrealign8(input):
    # batch channel time height width
    realign_input = torch.cuda.FloatTensor(input.shape[0], int(input.shape[1]/8), int(input.shape[2]*2), int(input.shape[3]*2), int(input.shape[4]*2))

    realign_input[:,:,0::2,0::2, 0::2] = input[:,0,:,:,:]
    realign_input[:,:,0::2,0::2, 1::2] = input[:,1,:,:,:]
    realign_input[:,:,0::2,1::2, 0::2] = input[:,2,:,:,:]
    realign_input[:,:,0::2,1::2, 1::2] = input[:,3,:,:,:]

    realign_input[:,:,1::2,0::2, 0::2] = input[:,4,:,:,:]
    realign_input[:,:,1::2,0::2, 1::2] = input[:,5,:,:,:]
    realign_input[:,:,1::2,1::2, 0::2] = input[:,6,:,:,:]
    realign_input[:,:,1::2,1::2, 1::2] = input[:,7,:,:,:]
    return realign_input


import argparse
import scipy.io as scio
import os
import time
import datetime
from skimage import io

from MN.utils import list2contours, Joint_Mask_List_Simple_Merge_Mask, Joint_Mask_List_Mul, Joint_Mask_List_Simple
from MN.utils import listAddcontours_Laplacian_pytorch, list2contours, listAddcontours_Laplacian_pytorch_Merge_Mask
############################################################################################################################################# 
parser = argparse.ArgumentParser()
parser.add_argument('--datasets_path', type=str, default='datasets', help="")
parser.add_argument('--datasets_folder1', type=str, default='test', help="")
parser.add_argument('--datasets_folder2', type=str, default='test', help="")
parser.add_argument('--output_dir', type=str, default='./results_mat', help="output directory")
opt = parser.parse_args()
############################################################################################################################################# 
if not os.path.exists(opt.output_dir): 
    os.mkdir(opt.output_dir)

current_time = datetime.datetime.now().strftime("%Y%m%d-%H%M")
SEG_output_path = opt.output_dir + '//' + 'R_'+opt.datasets_folder1+'_'+opt.datasets_folder2+'_'+current_time
if not os.path.exists(SEG_output_path): 
    os.mkdir(SEG_output_path)
SEG_f_con_output_path = SEG_output_path+ '//' + 'f_con'
if not os.path.exists(SEG_f_con_output_path): 
    os.mkdir(SEG_f_con_output_path)
SEG_mat_con_output_path = SEG_output_path+ '//' + 'mat'
if not os.path.exists(SEG_mat_con_output_path): 
    os.mkdir(SEG_mat_con_output_path)


im_folder1 = opt.datasets_path+'//'+opt.datasets_folder1
im_folder2 = opt.datasets_path+'//'+opt.datasets_folder2

for im_name1 in list(os.walk(im_folder1, topdown=False))[-1][-1]:
    data1 = scio.loadmat(im_folder1+'//'+im_name1)
    final_mask1 = data1['final_mask_list']

for im_name2 in list(os.walk(im_folder2, topdown=False))[-1][-1]:
    data2 = scio.loadmat(im_folder2+'//'+im_name2)
    final_mask2 = data2['final_mask_list']

# print(len(final_mask_list1[0]))
# print(len(final_mask_list2[0]))

final_mask_list1 = [] #1): #
for i in range(0,len(final_mask1[0])):
    neuron = final_mask1[0][i]
    neuron_centroid = neuron['centroid'][0][0][0]
    neuron_position = neuron['position'][0][0]
    neuron_trace = neuron['trace'][0][0][0]

    # print('neuron_position.shape[0]',neuron_position.shape[0])
    neuron_position_list =[] #1): #
    for p_i in range(0, neuron_position.shape[0]):
        now_position = list(neuron_position[p_i,:])
        neuron_position_list.append(now_position)
        # print(now_position[0], now_position[1])

    new_neuron = {}
    new_neuron['centroid'] = neuron_centroid
    new_neuron['position'] = neuron_position_list
    new_neuron['trace'] = neuron_trace

    # print(neuron_centroid)
    # print(neuron_trace)
    # print(neuron_position)
    '''
    for p_i in range(0,1): # len(position)):
        now_position = neuron_position[p_i]
        print(now_position[0], now_position[1])
    '''

    # print(neuron_position)
    # print(neuron_centroid[0])
    # print(neuron_centroid[0][0])
    # print(neuron_centroid[0][0][0])
    final_mask_list1.append(new_neuron)


final_mask_list2 = []
for i in range(0,len(final_mask2[0])):
    neuron = final_mask2[0][i]
    neuron_centroid = neuron['centroid'][0][0][0]
    neuron_position = neuron['position'][0][0]
    neuron_trace = neuron['trace'][0][0][0]

    # print('neuron_position.shape[0]',neuron_position.shape[0])
    neuron_position_list =[] #1): #
    for p_i in range(0, neuron_position.shape[0]):
        now_position = list(neuron_position[p_i,:])
        neuron_position_list.append(now_position)
        # print(now_position)

    new_neuron = {}
    new_neuron['centroid'] = neuron_centroid
    new_neuron['position'] = neuron_position_list
    new_neuron['trace'] = neuron_trace
    # print(neuron['centroid'])
    final_mask_list2.append(new_neuron)

print(len(final_mask_list1))
print(len(final_mask_list2))
# final_mask_list1[0]+final_mask_list2[0]

# final_mask_list = Joint_Mask_List_Simple_Merge_Mask(final_mask_list, [], corr_mark=0.7, area_mark=0.9, active_rate=0, if_coor=False, if_area=True, if_merge=False)
final_mask_list = Joint_Mask_List_Simple(final_mask_list1, final_mask_list2, corr_mark=0.7, area_mark=0.7, active_rate=0, if_coor=False, if_area=True, if_merge=True)

print(len(final_mask_list))

final_mask_list = listAddcontours_Laplacian_pytorch(final_mask_list, 2160, 2560)
final_contours,whole_contours = list2contours(final_mask_list, 2160, 2560)

img_f_contours_name = SEG_f_con_output_path+'//'+opt.datasets_folder1+'_'+opt.datasets_folder2+'_f_con.tif'
final_contours = final_contours.astype('uint16')
io.imsave(img_f_contours_name, final_contours)


mat_save_name = SEG_mat_con_output_path+'//'+opt.datasets_folder1+'_'+opt.datasets_folder2+'.mat'
data = {'a':final_mask_list, 'final_contours':final_contours}
scio.savemat(mat_save_name, {'final_mask_list':data['a'], 'final_contours':data['final_contours']})
'''
'''


import os
import torch
import torch.nn as nn
from torch.autograd import Variable
from torch.utils.data import DataLoader
import argparse
import time
import datetime
from skimage import io
import numpy as np

from deepwonder.RMBG.data_process import test_preprocess_lessMemoryNoTail_SubImg, testset, multibatch_test_save, singlebatch_test_save
from deepwonder.RMBG.network import Network_3D_Unet
from deepwonder.RMBG.utils import FFDrealign4, inv_FFDrealign4
#############################################################################################################################################
def wf2nobg_ffd(net,
                sub_img,
                RMBG_GPU='0', 
                RMBG_batch_size=1, 
                RMBG_img_w=256, 
                RMBG_img_h=256, 
                RMBG_img_s=256, 
                RMBG_gap_w=224, 
                RMBG_gap_h=224, 
                RMBG_gap_s=224, 
                RMBG_normalize_factor=1):
    #############################################################################################################################################
    opt = {}
    #############################################################################################################################################
    opt['GPU'] = RMBG_GPU
    opt['batch_size'] = RMBG_batch_size
    opt['img_w'] = RMBG_img_w
    opt['img_h'] = RMBG_img_h
    opt['img_s'] = RMBG_img_s

    opt['gap_w'] = RMBG_gap_w
    opt['gap_h'] = RMBG_gap_h
    opt['gap_s'] = RMBG_gap_s

    opt['normalize_factor'] = RMBG_normalize_factor
    #############################################################################################################################################
    os.environ["CUDA_VISIBLE_DEVICES"] = str(opt['GPU'])

    ##############################################################################################################################################################
    if torch.cuda.is_available():
        net = net.cuda()
    cuda = True if torch.cuda.is_available() else False
    Tensor = torch.cuda.FloatTensor if cuda else torch.FloatTensor
    ##############################################################################################################################################################
    name_list, noise_img, coordinate_list = test_preprocess_lessMemoryNoTail_SubImg(opt, sub_img)

    #print(len(name_list))
    prev_time = time.time()
    time_start = time.time()
    denoise_img = np.zeros(noise_img.shape)
    input_img = np.zeros(noise_img.shape)

    test_data = testset(name_list, coordinate_list, noise_img)
    testloader = DataLoader(test_data, batch_size=opt['batch_size'], shuffle=False, num_workers=4)
    for iteration, (noise_patch,single_coordinate) in enumerate(testloader):
        noise_patch=noise_patch.cuda().float()
        # print('noise_patch -----> ',noise_patch.shape)
        real_A = FFDrealign4(noise_patch)
        real_A = Variable(real_A)
        
        # print('real_A -----> ',real_A.shape)
        fake_B = net(real_A)
        ################################################################################################################
        # Determine approximate time left
        batches_done = iteration
        batches_left = 1 * len(testloader) - batches_done
        time_left_seconds = int(batches_left * (time.time() - prev_time))
        time_left = datetime.timedelta(seconds=time_left_seconds)
        prev_time = time.time()
        ################################################################################################################
        if iteration % 1 == 0:
            time_end = time.time()
            time_cost = time_end - time_start  # datetime.timedelta(seconds= (time_end - time_start))
            print(
                '\r\033[1;32m[RMBG]\033[0m [Patch %d/%d] [Time Cost: %.0d s] [ETA: %s s]     '
                % ( iteration + 1,
                    len(testloader),
                    time_cost,
                    time_left_seconds
                ), end=' ')

        if (iteration + 1) % len(testloader) == 0:
            print('\n', end=' ')
        ################################################################################################################
        fake_B_realign = inv_FFDrealign4(fake_B)
        output_image = np.squeeze(fake_B_realign.cpu().detach().numpy())
        real_A_realign = inv_FFDrealign4(real_A)
        raw_image = np.squeeze(real_A_realign.cpu().detach().numpy())
        if(output_image.ndim==3):
            turn=1
        else:
            turn=output_image.shape[0]
        #print(turn)
        if(turn>1):
            for id in range(turn):
                #print('shape of output_image -----> ',output_image.shape)
                aaaa,bbbb,stack_start_w,stack_end_w,stack_start_h,stack_end_h,stack_start_s,stack_end_s=multibatch_test_save(single_coordinate,id,output_image,raw_image)
                denoise_img[stack_start_s:stack_end_s, stack_start_h:stack_end_h, stack_start_w:stack_end_w] \
                = aaaa 
                input_img[stack_start_s:stack_end_s, stack_start_h:stack_end_h, stack_start_w:stack_end_w] \
                = bbbb
        else:
            aaaa, bbbb, stack_start_w, stack_end_w, stack_start_h, stack_end_h, stack_start_s, stack_end_s = singlebatch_test_save(
                single_coordinate, output_image, raw_image)
            denoise_img[stack_start_s:stack_end_s, stack_start_h:stack_end_h, stack_start_w:stack_end_w] \
                            = aaaa 
            input_img[stack_start_s:stack_end_s, stack_start_h:stack_end_h, stack_start_w:stack_end_w] \
                = bbbb
            

    # del noise_img
    output_img = denoise_img.squeeze().astype(np.float32) * opt['normalize_factor']
    del denoise_img
    # print(np.min(output_img),' -----> ',np.max(output_img))
    output_img = np.clip(output_img, 0, 65535).astype('uint16')
    # output_img = output_img - output_img.min()
    # output_img = output_img.astype('uint16')

    return output_img




import os
import sys
import io

import numpy as np
import torch
import cv2
from skimage import io

########################################################################################################################
def create_feature_maps(init_channel_number, number_of_fmaps):
    return [init_channel_number * 2 ** k for k in range(number_of_fmaps)]


def name2index(opt, input_name, num_h, num_w, num_s):
    # print(input_name)
    name_list = input_name.split('_')
    # print(name_list)
    z_part = name_list[-1]
    # print(z_part)
    y_part = name_list[-2]
    # print(y_part)
    x_part = name_list[-3]
    # print(x_part)
    z_index = int(z_part.replace('z',''))
    y_index = int(y_part.replace('y',''))
    x_index = int(x_part.replace('x',''))
    # print("x_index ---> ",x_index,"y_index ---> ", y_index,"z_index ---> ", z_index)

    cut_w = (opt.img_w - opt.gap_w)/2
    cut_h = (opt.img_h - opt.gap_h)/2
    cut_s = (opt.img_s - opt.gap_s)/2
    # print("z_index ---> ",cut_w, "cut_h ---> ",cut_h, "cut_s ---> ",cut_s)
    if x_index == 0:
        stack_start_w = x_index*opt.gap_w
        stack_end_w = x_index*opt.gap_w+opt.img_w-cut_w
        patch_start_w = 0
        patch_end_w = opt.img_w-cut_w
    elif x_index == num_w-1:
        stack_start_w = x_index*opt.gap_w+cut_w
        stack_end_w = x_index*opt.gap_w+opt.img_w
        patch_start_w = cut_w
        patch_end_w = opt.img_w
    else:
        stack_start_w = x_index*opt.gap_w+cut_w
        stack_end_w = x_index*opt.gap_w+opt.img_w-cut_w
        patch_start_w = cut_w
        patch_end_w = opt.img_w-cut_w

    if y_index == 0:
        stack_start_h = y_index*opt.gap_h
        stack_end_h = y_index*opt.gap_h+opt.img_h-cut_h
        patch_start_h = 0
        patch_end_h = opt.img_h-cut_h
    elif y_index == num_h-1:
        stack_start_h = y_index*opt.gap_h+cut_h
        stack_end_h = y_index*opt.gap_h+opt.img_h
        patch_start_h = cut_h
        patch_end_h = opt.img_h
    else:
        stack_start_h = y_index*opt.gap_h+cut_h
        stack_end_h = y_index*opt.gap_h+opt.img_h-cut_h
        patch_start_h = cut_h
        patch_end_h = opt.img_h-cut_h

    if z_index == 0:
        stack_start_s = z_index*opt.gap_s
        stack_end_s = z_index*opt.gap_s+opt.img_s-cut_s
        patch_start_s = 0
        patch_end_s = opt.img_s-cut_s
    elif z_index == num_s-1:
        stack_start_s = z_index*opt.gap_s+cut_s
        stack_end_s = z_index*opt.gap_s+opt.img_s
        patch_start_s = cut_s
        patch_end_s = opt.img_s
    else:
        stack_start_s = z_index*opt.gap_s+cut_s
        stack_end_s = z_index*opt.gap_s+opt.img_s-cut_s
        patch_start_s = cut_s
        patch_end_s = opt.img_s-cut_s
    return int(stack_start_w) ,int(stack_end_w) ,int(patch_start_w) ,int(patch_end_w) ,\
    int(stack_start_h) ,int(stack_end_h) ,int(patch_start_h) ,int(patch_end_h), \
    int(stack_start_s) ,int(stack_end_s) ,int(patch_start_s) ,int(patch_end_s)

def save_tiff_image(args, image_tensor, image_path):
    image = (image_tensor.cpu().detach().numpy()*args.normalize_factor) #.astype(np.uint16)
    # print('image max ',np.max(image),' image min ',np.min(image))
    '''
    image_tensor = torch.from_numpy(image_numpy)
    image_new_numpy = image_tensor.permute(2, 0, 1).numpy()
    # image_new_numpy = image_numpy
    if '_A' in image_path:
        image_new_numpy = image_new_numpy[...,np.newaxis] 
    '''
    save_tiff_path = image_path.replace('.png','.tif')
    io.imsave(save_tiff_path,image)

def save_numpy_image(args, image_tensor, image_path):
    image = (image_tensor*args.normalize_factor) #.astype(np.uint16)
    save_tiff_path = image_path.replace('.png','.tif')
    io.imsave(save_tiff_path,image)

def save_feature_tiff_image(image_tensor, image_path):
    image = image_tensor.cpu().detach().numpy()*255 #).astype(np.uint8)
    # image = np.clip(image, 0, 65535).astype('uint16')
    image = np.clip(image, 0, 255).astype('uint8')
    # print('image max ',np.max(image),' image min ',np.min(image))
    save_tiff_path = image_path.replace('.png','.tif')
    io.imsave(save_tiff_path,image)


def FFDrealign4(input):
    # batch channel time height width
    realign_input = torch.cuda.FloatTensor(input.shape[0], input.shape[1]*4, input.shape[2], int(input.shape[3]/2), int(input.shape[4]/2))
    # print('realign_input -----> ',realign_input.shape)
    # print('input -----> ',input.shape)
    # print('realign_input[:,0,:,:,:] -----> ',realign_input[:,0,:,:,:].shape)
    # print('input[:,:,:,0::2, 0::2] -----> ',input[:,:,:,0::2, 0::2].shape)
    realign_input[:,0,:,:,:] = torch.squeeze(input[:,:,:,0::2, 0::2])
    realign_input[:,1,:,:,:] = torch.squeeze(input[:,:,:,0::2, 1::2])
    realign_input[:,2,:,:,:] = torch.squeeze(input[:,:,:,1::2, 0::2])
    realign_input[:,3,:,:,:] = torch.squeeze(input[:,:,:,1::2, 1::2])
    return realign_input

def inv_FFDrealign4(input):
    # batch channel time height width
    realign_input = torch.cuda.FloatTensor(input.shape[0], int(input.shape[1]/4), input.shape[2], int(input.shape[3]*2), int(input.shape[4]*2))

    # print('realign_input[:,:,:,0::2, 0::2] -----> ',realign_input[:,:,:,0::2, 0::2].shape)
    # print('input[:,0,:,:,:] -----> ',input[:,0,:,:,:].shape)
    realign_input[:,:,:,0::2, 0::2] = torch.unsqueeze(input[:,0,:,:,:], 1)
    realign_input[:,:,:,0::2, 1::2] = torch.unsqueeze(input[:,1,:,:,:], 1)
    realign_input[:,:,:,1::2, 0::2] = torch.unsqueeze(input[:,2,:,:,:], 1)
    realign_input[:,:,:,1::2, 1::2] = torch.unsqueeze(input[:,3,:,:,:], 1)
    return realign_input


import numpy as np
import os
import tifffile as tiff
import random
import math
import torch
from torch.utils.data import Dataset


def Img2Subimg(args):

    im_folder = args.datasets_path+'//'+args.datasets_folder

    name_list = []
    # train_raw = []
    image_list={}
    coordinate_list={}
    print('im_folder ---> ',im_folder)
    print('list(os.walk(im_folder, topdown=False))[-1][-1] ---> ',list(os.walk(im_folder, topdown=False))[-1][-1])
    for im_name in list(os.walk(im_folder, topdown=False))[-1][-1]:
        img_h = args.sub_img_h
        img_w = args.sub_img_w
        img_s2 = args.sub_img_s
        gap_h = args.sub_gap_h
        gap_w = args.sub_gap_w
        gap_s2 = args.sub_gap_s
        cut_w = (img_w - gap_w)/2
        cut_h = (img_h - gap_h)/2
        cut_s = (img_s2 - gap_s2)/2

        print('im_name -----> ',im_name)
        im_dir = im_folder+'//'+im_name
        im = tiff.imread(im_dir)
        print('im -----> ',im.shape)

        if '.tiff' in im_name:
            im_name = im_name.replace('.tiff','')
        if ('.tif' in im_name)&('.tiff' not in im_name):
            im_name = im_name.replace('.tif','')
        name_list.append(im_name)

        if im.shape[0]>args.test_datasize:
            im = im[0:args.test_datasize,:,:].copy()
        if im.shape[0]<img_s2:
            img_s2 = im.shape[0]
            args.sub_img_s = img_s2
            gap_s2 = img_s2
            cut_s = 0
        if im.shape[1]<img_h:
            img_h = im.shape[1]
            args.sub_img_h = img_h
            gap_h = img_h
            cut_h = 0
        if im.shape[2]<img_w:
            img_w = im.shape[2]
            args.sub_img_w = img_w
            gap_w = img_w
            cut_w = 0
        
        image_list[im_name.replace('.tif','')] = im

        whole_w = im.shape[2]
        whole_h = im.shape[1]
        whole_s = im.shape[0]

        num_w = math.ceil((whole_w-img_w+gap_w)/gap_w)
        num_h = math.ceil((whole_h-img_h+gap_h)/gap_h)
        num_s = math.ceil((whole_s-img_s2+gap_s2)/gap_s2)

        per_coor_list = []
        for x in range(0,num_h):
            for y in range(0,num_w):
                for z in range(0,num_s):
                    # print('x -----> ',x,num_h,'y -----> ',y,num_w,'z -----> ',z,num_s)
                    per_coor = {}
                    if x != (num_h-1):
                        init_h = gap_h*x
                        end_h = gap_h*x + img_h
                    elif x == (num_h-1):
                        init_h = whole_h - img_h
                        end_h = whole_h

                    if y != (num_w-1):
                        init_w = gap_w*y
                        end_w = gap_w*y + img_w
                    elif y == (num_w-1):
                        init_w = whole_w - img_w
                        end_w = whole_w

                    if z != (num_s-1):
                        init_s = gap_s2*z
                        end_s = gap_s2*z + img_s2
                    elif z == (num_s-1):
                        init_s = whole_s - img_s2
                        end_s = whole_s
                    per_coor['init_h'] = init_h
                    per_coor['end_h'] = end_h
                    per_coor['init_w'] = init_w
                    per_coor['end_w'] = end_w
                    per_coor['init_s'] = init_s
                    per_coor['end_s'] = end_s

                    if (num_w-1)==0:
                        per_coor['stack_start_w'] = 0
                        per_coor['stack_end_w'] = whole_w
                        per_coor['patch_start_w'] = 0
                        per_coor['patch_end_w'] = img_w
                    elif (num_w-1)>0:
                        if y == 0:
                            per_coor['stack_start_w'] = y*gap_w
                            per_coor['stack_end_w'] = y*gap_w+img_w-cut_w
                            per_coor['patch_start_w'] = 0
                            per_coor['patch_end_w'] = img_w-cut_w
                        elif y == num_w-1:
                            per_coor['stack_start_w'] = whole_w-img_w+cut_w
                            per_coor['stack_end_w'] = whole_w
                            per_coor['patch_start_w'] = cut_w
                            per_coor['patch_end_w'] = img_w
                        else:
                            per_coor['stack_start_w'] = y*gap_w+cut_w
                            per_coor['stack_end_w'] = y*gap_w+img_w-cut_w
                            per_coor['patch_start_w'] = cut_w
                            per_coor['patch_end_w'] = img_w-cut_w

                    if (num_h-1)==0:
                        per_coor['stack_start_h'] = 0
                        per_coor['stack_end_h'] = whole_h
                        per_coor['patch_start_h'] = 0
                        per_coor['patch_end_h'] = img_h
                    elif (num_h-1)>0:
                        if x == 0:
                            per_coor['stack_start_h'] = x*gap_h
                            per_coor['stack_end_h'] = x*gap_h+img_h-cut_h
                            per_coor['patch_start_h'] = 0
                            per_coor['patch_end_h'] = img_h-cut_h
                        elif x == num_h-1:
                            per_coor['stack_start_h'] = whole_h-img_h+cut_h
                            per_coor['stack_end_h'] = whole_h
                            per_coor['patch_start_h'] = cut_h
                            per_coor['patch_end_h'] = img_h
                        else:
                            per_coor['stack_start_h'] = x*gap_h+cut_h
                            per_coor['stack_end_h'] = x*gap_h+img_h-cut_h
                            per_coor['patch_start_h'] = cut_h
                            per_coor['patch_end_h'] = img_h-cut_h

                    if (num_s-1)==0:
                        per_coor['stack_start_s'] = 0
                        per_coor['stack_end_s'] = whole_s
                        per_coor['patch_start_s'] = 0
                        per_coor['patch_end_s'] = img_s2 
                    elif (num_s-1)>0:             
                        if z == 0:
                            per_coor['stack_start_s'] = z*gap_s2
                            per_coor['stack_end_s'] = z*gap_s2+img_s2-cut_s
                            per_coor['patch_start_s'] = 0
                            per_coor['patch_end_s'] = img_s2-cut_s
                        elif z == num_s-1:
                            per_coor['stack_start_s'] = whole_s-img_s2+cut_s
                            per_coor['stack_end_s'] = whole_s
                            per_coor['patch_start_s'] = cut_s
                            per_coor['patch_end_s'] = img_s2
                        else:
                            per_coor['stack_start_s'] = z*gap_s2+cut_s
                            per_coor['stack_end_s'] = z*gap_s2+img_s2-cut_s
                            per_coor['patch_start_s'] = cut_s
                            per_coor['patch_end_s'] = img_s2-cut_s
                    
                    per_coor['z'] = z

                    # noise_patch1 = noise_im[init_s:end_s,init_h:end_h,init_w:end_w]
                    patch_name = args.datasets_folder+'_x'+str(x)+'_y'+str(y)+'_z'+str(z)
                    per_coor['name'] = im_name.replace('.tif','')
                    # print(' single_coordinate -----> ',single_coordinate)
                    per_coor_list.append(per_coor)
        # print(' per_coor_list -----> ',len(per_coor_list))
        '''
        while len(per_coor_list)%args.batch_size!=0:
            per_coor_list.append(per_coor)
            print(' per_coor_list -----> ',len(per_coor_list),' -----> ',len(per_coor_list)%args.batch_size,' -----> ',args.batch_size)
        '''
        coordinate_list[im_name.replace('.tif','')] = per_coor_list
    return  name_list, image_list, coordinate_list


import os
import torch
import torch.nn as nn
from torch.autograd import Variable
from torch.utils.data import DataLoader
import argparse
import time
import datetime
from RMBG.network import Network_3D_Unet
import numpy as np

from RMBG.data_process import test_preprocess_lessMemoryNoTail_SubImg, testset, multibatch_test_save, singlebatch_test_save
from skimage import io
from RMBG.utils import FFDrealign8, inv_FFDrealign8
#############################################################################################################################################
def wf2nobg_ffd8(net,
                sub_img,
                RMBG_GPU='0', 
                RMBG_batch_size=1, 
                RMBG_img_w=256, 
                RMBG_img_h=256, 
                RMBG_img_s=256, 
                RMBG_gap_w=224, 
                RMBG_gap_h=224, 
                RMBG_gap_s=224, 
                RMBG_normalize_factor=1):
    #############################################################################################################################################
    opt = {}
    #############################################################################################################################################
    opt['GPU'] = RMBG_GPU
    opt['batch_size'] = RMBG_batch_size
    opt['img_w'] = RMBG_img_w
    opt['img_h'] = RMBG_img_h
    opt['img_s'] = RMBG_img_s

    opt['gap_w'] = RMBG_gap_w
    opt['gap_h'] = RMBG_gap_h
    opt['gap_s'] = RMBG_gap_s

    opt['normalize_factor'] = RMBG_normalize_factor
    #############################################################################################################################################
    os.environ["CUDA_VISIBLE_DEVICES"] = str(opt['GPU'])

    ##############################################################################################################################################################
    if torch.cuda.is_available():
        net = net.cuda()
    cuda = True if torch.cuda.is_available() else False
    Tensor = torch.cuda.FloatTensor if cuda else torch.FloatTensor
    ##############################################################################################################################################################
    name_list, noise_img, coordinate_list = test_preprocess_lessMemoryNoTail_SubImg(opt, sub_img)

    #print(len(name_list))
    prev_time = time.time()
    time_start = time.time()
    denoise_img = np.zeros(noise_img.shape)
    input_img = np.zeros(noise_img.shape)

    test_data = testset(name_list, coordinate_list, noise_img)
    testloader = DataLoader(test_data, batch_size=opt['batch_size'], shuffle=False, num_workers=4)
    for iteration, (noise_patch,single_coordinate) in enumerate(testloader):
        noise_patch=noise_patch.cuda().float()
        # print('noise_patch -----> ',noise_patch.shape)
        real_A = FFDrealign8(noise_patch)
        real_A = Variable(real_A)
        
        # print('real_A -----> ',real_A.shape)
        fake_B = net(real_A)
        # print('fake_B -----> ',fake_B.shape)
        ################################################################################################################
        # Determine approximate time left
        batches_done = iteration
        batches_left = 1 * len(testloader) - batches_done
        time_left_seconds = int(batches_left * (time.time() - prev_time))
        time_left = datetime.timedelta(seconds=time_left_seconds)
        prev_time = time.time()
        ################################################################################################################
        if iteration % 1 == 0:
            time_end = time.time()
            time_cost = time_end - time_start  # datetime.timedelta(seconds= (time_end - time_start))
            print(
                '\r\033[1;32m[RMBG]\033[0m [Patch %d/%d] [Time Cost: %.0d s] [ETA: %s s]     '
                % ( iteration + 1,
                    len(testloader),
                    time_cost,
                    time_left_seconds
                ), end=' ')

        if (iteration + 1) % len(testloader) == 0:
            print('\n', end=' ')
        ################################################################################################################
        fake_B_realign = inv_FFDrealign8(fake_B)
        output_image = np.squeeze(fake_B_realign.cpu().detach().numpy())
        # print('output_image',output_image.shape)
        real_A_realign = inv_FFDrealign8(real_A)
        raw_image = np.squeeze(real_A_realign.cpu().detach().numpy())
        if(output_image.ndim==3):
            turn=1
        else:
            turn=output_image.shape[0]
        #print(turn)
        if(turn>1):
            for id in range(turn):
                #print('shape of output_image -----> ',output_image.shape)
                aaaa,bbbb,stack_start_w,stack_end_w,stack_start_h,stack_end_h,stack_start_s,stack_end_s=multibatch_test_save(single_coordinate,id,output_image,raw_image)
                denoise_img[stack_start_s:stack_end_s, stack_start_h:stack_end_h, stack_start_w:stack_end_w] \
                = aaaa 
                input_img[stack_start_s:stack_end_s, stack_start_h:stack_end_h, stack_start_w:stack_end_w] \
                = bbbb
        else:
            aaaa, bbbb, stack_start_w, stack_end_w, stack_start_h, stack_end_h, stack_start_s, stack_end_s = singlebatch_test_save(
                single_coordinate, output_image, raw_image)
            denoise_img[stack_start_s:stack_end_s, stack_start_h:stack_end_h, stack_start_w:stack_end_w] \
                            = aaaa 
            input_img[stack_start_s:stack_end_s, stack_start_h:stack_end_h, stack_start_w:stack_end_w] \
                = bbbb
            

    # del noise_img
    output_img = denoise_img.squeeze().astype(np.float32) * opt['normalize_factor']
    del denoise_img
    # print(np.min(output_img),' -----> ',np.max(output_img))
    output_img = np.clip(output_img, 0, 65535).astype('uint16')
    # output_img = output_img - output_img.min()
    # output_img = output_img.astype('uint16')

    return output_img




import numpy as np
import os
import tifffile as tiff
import random
import math
import torch
from torch.utils.data import Dataset

###
#  this file is for pre-processing images for training and test.
#  currently we use splitted dataset


###############################
# dataset class

class trainset(Dataset):
    def __init__(self,name_list,coordinate_list,noise_img, noise_img2):
        self.name_list = name_list
        self.coordinate_list=coordinate_list
        self.noise_img = noise_img
        self.noise_img2 = noise_img2

    def __getitem__(self, index):
        #fn = self.images[index]
        single_coordinate = self.coordinate_list[self.name_list[index]]
        init_h = single_coordinate['init_h']
        end_h = single_coordinate['end_h']
        init_w = single_coordinate['init_w']
        end_w = single_coordinate['end_w']
        init_s = single_coordinate['init_s']
        end_s = single_coordinate['end_s']
        input = self.noise_img[init_s: end_s : 1, init_h:end_h, init_w:end_w]
        target = self.noise_img2[init_s: end_s : 1, init_h:end_h, init_w:end_w]

        input=torch.from_numpy(np.expand_dims(input, 0))
        target=torch.from_numpy(np.expand_dims(target, 0))
        #target = self.target[index]
        return input, target

    def __len__(self):
        return len(self.name_list)


class trainsetMul(Dataset):
    def __init__(self,coor_list, GT_list, input_list, opt):
        self.coor_list = coor_list
        self.GT_list = GT_list
        self.input_list = input_list
        self.opt = opt

    def __getitem__(self, index):
        #fn = self.images[index]
        per_coor = self.coor_list[index]
        train_im_name = per_coor['name']
        init_w = per_coor['init_w']
        init_h = per_coor['init_h']
        init_s = per_coor['init_s']
        GT_im = self.GT_list[train_im_name]
        input_im = self.input_list[train_im_name]
        
        GT_patch =       GT_im[init_s:init_s+self.opt.img_s, init_w:init_w+self.opt.img_w, init_h:init_h+self.opt.img_h]
        input_patch = input_im[init_s:init_s+self.opt.img_s, init_w:init_w+self.opt.img_w, init_h:init_h+self.opt.img_h]

        # input = self.noise_img[init_s: end_s : 1, init_h:end_h, init_w:end_w]
        # target = self.noise_img2[init_s: end_s : 1, init_h:end_h, init_w:end_w]

        input=torch.from_numpy(np.expand_dims(input_patch, 0))
        target=torch.from_numpy(np.expand_dims(GT_patch, 0))
        #target = self.target[index]
        return input, target

    def __len__(self):
        return len(self.coor_list)


class testset(Dataset):
    def __init__(self,name_list,coordinate_list,noise_img):
        self.name_list = name_list
        self.coordinate_list=coordinate_list
        self.noise_img = noise_img

    def __getitem__(self, index):
        #fn = self.images[index]
        single_coordinate = self.coordinate_list[self.name_list[index]]
        init_h = single_coordinate['init_h']
        end_h = single_coordinate['end_h']
        init_w = single_coordinate['init_w']
        end_w = single_coordinate['end_w']
        init_s = single_coordinate['init_s']
        end_s = single_coordinate['end_s']
        noise_patch = self.noise_img[init_s:end_s, init_h:end_h, init_w:end_w]
        noise_patch=torch.from_numpy(np.expand_dims(noise_patch, 0))
        #target = self.target[index]
        return noise_patch,single_coordinate

    def __len__(self):
        return len(self.name_list)
###############################
# save
def singlebatch_test_save(single_coordinate,output_image,raw_image):
    stack_start_w = int(single_coordinate['stack_start_w'])
    stack_end_w = int(single_coordinate['stack_end_w'])
    patch_start_w = int(single_coordinate['patch_start_w'])
    patch_end_w = int(single_coordinate['patch_end_w'])

    stack_start_h = int(single_coordinate['stack_start_h'])
    stack_end_h = int(single_coordinate['stack_end_h'])
    patch_start_h = int(single_coordinate['patch_start_h'])
    patch_end_h = int(single_coordinate['patch_end_h'])

    stack_start_s = int(single_coordinate['stack_start_s'])
    stack_end_s = int(single_coordinate['stack_end_s'])
    patch_start_s = int(single_coordinate['patch_start_s'])
    patch_end_s = int(single_coordinate['patch_end_s'])
    '''
    print('RMBG stack_start_w -----> ',stack_start_w, ' patch_start_w -----> ', \
    patch_start_w,'stack_end_w -----> ',stack_end_w, ' patch_end_w -----> ',patch_end_w)
    print('RMBG stack_start_h -----> ',stack_start_h, ' patch_start_h -----> ', \
    patch_start_h, 'stack_end_h -----> ',stack_end_h, ' patch_end_h -----> ',patch_end_h)
    '''
    aaaa = output_image[patch_start_s:patch_end_s, patch_start_h:patch_end_h, patch_start_w:patch_end_w]
    bbbb = raw_image[patch_start_s:patch_end_s, patch_start_h:patch_end_h, patch_start_w:patch_end_w]
    return aaaa,bbbb,stack_start_w,stack_end_w,stack_start_h,stack_end_h,stack_start_s,stack_end_s


def multibatch_test_save(single_coordinate,id,output_image,raw_image):
    stack_start_w_id = single_coordinate['stack_start_w'].numpy()
    stack_start_w = int(stack_start_w_id[id])
    stack_end_w_id = single_coordinate['stack_end_w'].numpy()
    stack_end_w=int(stack_end_w_id[id])
    patch_start_w_id = single_coordinate['patch_start_w'].numpy()
    patch_start_w=int(patch_start_w_id[id])
    patch_end_w_id = single_coordinate['patch_end_w'].numpy()
    patch_end_w=int(patch_end_w_id[id])

    stack_start_h_id = single_coordinate['stack_start_h'].numpy()
    stack_start_h = int(stack_start_h_id[id])
    stack_end_h_id = single_coordinate['stack_end_h'].numpy()
    stack_end_h = int(stack_end_h_id[id])
    patch_start_h_id = single_coordinate['patch_start_h'].numpy()
    patch_start_h = int(patch_start_h_id[id])
    patch_end_h_id = single_coordinate['patch_end_h'].numpy()
    patch_end_h = int(patch_end_h_id[id])

    stack_start_s_id = single_coordinate['stack_start_s'].numpy()
    stack_start_s = int(stack_start_s_id[id])
    stack_end_s_id = single_coordinate['stack_end_s'].numpy()
    stack_end_s = int(stack_end_s_id[id])
    patch_start_s_id = single_coordinate['patch_start_s'].numpy()
    patch_start_s = int(patch_start_s_id[id])
    patch_end_s_id = single_coordinate['patch_end_s'].numpy()
    patch_end_s = int(patch_end_s_id[id])

    output_image_id=output_image[id]
    raw_image_id=raw_image[id]
    aaaa = output_image_id[patch_start_s:patch_end_s, patch_start_h:patch_end_h, patch_start_w:patch_end_w]
    bbbb = raw_image_id[patch_start_s:patch_end_s, patch_start_h:patch_end_h, patch_start_w:patch_end_w]

    return aaaa,bbbb,stack_start_w,stack_end_w,stack_start_h,stack_end_h,stack_start_s,stack_end_s

###############################
# shuffle
def shuffle_datasets(train_raw, train_GT, name_list):
    # target: train_raw, train_GT, name_list
    index_list = list(range(0, len(name_list)))
    # print('index_list -----> ',index_list)
    random.shuffle(index_list)
    random_index_list = index_list
    # print('index_list -----> ',index_list)
    new_name_list = list(range(0, len(name_list)))
    train_raw = np.array(train_raw)
    # print('train_raw shape -----> ',train_raw.shape)
    train_GT = np.array(train_GT)
    # print('train_GT shape -----> ',train_GT.shape)
    new_train_raw = train_raw
    new_train_GT = train_GT
    for i in range(0,len(random_index_list)):
        # print('i -----> ',i)
        new_train_raw[i,:,:,:] = train_raw[random_index_list[i],:,:,:] 
        new_train_GT[i,:,:,:] = train_GT[random_index_list[i],:,:,:]
        new_name_list[i] = name_list[random_index_list[i]]
    # new_train_raw = np.expand_dims(new_train_raw, 4)
    # new_train_GT = np.expand_dims(new_train_GT, 4)
    return new_train_raw, new_train_GT, new_name_list

###############################
# pre-process
def get_gap_s(args, img, stack_num):
    whole_w = img.shape[2]
    whole_h = img.shape[1]
    whole_s = img.shape[0]
    # print('whole_w -----> ',whole_w)
    # print('whole_h -----> ',whole_h)
    # print('whole_s -----> ',whole_s)
    w_num = math.floor((whole_w-args.img_w)/args.gap_w)+1
    h_num = math.floor((whole_h-args.img_h)/args.gap_h)+1
    s_num = math.ceil(args.train_datasets_size/w_num/h_num/stack_num)
    # print('w_num -----> ',w_num)
    # print('h_num -----> ',h_num)
    # print('s_num -----> ',s_num)
    gap_s = math.floor((whole_s-args.img_s*2)/(s_num-1))
    # print('gap_s -----> ',gap_s)
    return gap_s


def test_preprocess_lessMemoryNoTail_chooseOne (args, N):
    img_h = args.img_h
    img_w = args.img_w
    img_s2 = args.img_s
    gap_h = args.gap_h
    gap_w = args.gap_w
    gap_s2 = args.gap_s
    cut_w = (img_w - gap_w)/2
    cut_h = (img_h - gap_h)/2
    cut_s = (img_s2 - gap_s2)/2
    im_folder = args.datasets_path+'//'+args.datasets_folder

    name_list = []
    # train_raw = []
    coordinate_list={}
    img_list = list(os.walk(im_folder, topdown=False))[-1][-1]
    img_list.sort()
    # print(img_list)

    im_name = img_list[N]

    im_dir = im_folder+'//'+im_name
    print('im_dir -----> ',im_dir)
    noise_im = tiff.imread(im_dir)
    # print('noise_im shape -----> ',noise_im.shape)
    # print('noise_im max -----> ',noise_im.max())
    # print('noise_im min -----> ',noise_im.min())
    if noise_im.shape[0]>args.test_datasize:
        noise_im = noise_im[0:args.test_datasize,:,:]
    noise_im = (noise_im).astype(np.float32)/args.normalize_factor

    noise_im_ave_single = np.mean(noise_im, axis=0)
    noise_im_ave = np.zeros(noise_im.shape)
    for i in range(0, noise_im.shape[0]):
        noise_im_ave[i,:,:] = noise_im_ave_single
    noise_im = noise_im-noise_im_ave


    whole_w = noise_im.shape[2]
    whole_h = noise_im.shape[1]
    whole_s = noise_im.shape[0]

    num_w = math.ceil((whole_w-img_w+gap_w)/gap_w)
    num_h = math.ceil((whole_h-img_h+gap_h)/gap_h)
    num_s = math.ceil((whole_s-img_s2+gap_s2)/gap_s2)
    # print('int((whole_h-img_h+gap_h)/gap_h) -----> ',int((whole_h-img_h+gap_h)/gap_h))
    # print('int((whole_w-img_w+gap_w)/gap_w) -----> ',int((whole_w-img_w+gap_w)/gap_w))
    # print('int((whole_s-img_s2+gap_s2)/gap_s2) -----> ',int((whole_s-img_s2+gap_s2)/gap_s2))
    for x in range(0,num_h):
        for y in range(0,num_w):
            for z in range(0,num_s):
                single_coordinate={'init_h':0, 'end_h':0, 'init_w':0, 'end_w':0, 'init_s':0, 'end_s':0}
                if x != (num_h-1):
                    init_h = gap_h*x
                    end_h = gap_h*x + img_h
                elif x == (num_h-1):
                    init_h = whole_h - img_h
                    end_h = whole_h

                if y != (num_w-1):
                    init_w = gap_w*y
                    end_w = gap_w*y + img_w
                elif y == (num_w-1):
                    init_w = whole_w - img_w
                    end_w = whole_w

                if z != (num_s-1):
                    init_s = gap_s2*z
                    end_s = gap_s2*z + img_s2
                elif z == (num_s-1):
                    init_s = whole_s - img_s2
                    end_s = whole_s
                single_coordinate['init_h'] = init_h
                single_coordinate['end_h'] = end_h
                single_coordinate['init_w'] = init_w
                single_coordinate['end_w'] = end_w
                single_coordinate['init_s'] = init_s
                single_coordinate['end_s'] = end_s

                if (num_w-1)==0:
                    single_coordinate['stack_start_w'] = 0
                    single_coordinate['stack_end_w'] = whole_w
                    single_coordinate['patch_start_w'] = 0
                    single_coordinate['patch_end_w'] = img_w
                elif (num_w-1)>0:
                    if y == 0:
                        single_coordinate['stack_start_w'] = y*gap_w
                        single_coordinate['stack_end_w'] = y*gap_w+img_w-cut_w
                        single_coordinate['patch_start_w'] = 0
                        single_coordinate['patch_end_w'] = img_w-cut_w
                    elif y == num_w-1:
                        single_coordinate['stack_start_w'] = whole_w-img_w+cut_w
                        single_coordinate['stack_end_w'] = whole_w
                        single_coordinate['patch_start_w'] = cut_w
                        single_coordinate['patch_end_w'] = img_w
                    else:
                        single_coordinate['stack_start_w'] = y*gap_w+cut_w
                        single_coordinate['stack_end_w'] = y*gap_w+img_w-cut_w
                        single_coordinate['patch_start_w'] = cut_w
                        single_coordinate['patch_end_w'] = img_w-cut_w

                if (num_h-1)==0:
                    single_coordinate['stack_start_h'] = 0
                    single_coordinate['stack_end_h'] = whole_h
                    single_coordinate['patch_start_h'] = 0
                    single_coordinate['patch_end_h'] = img_h
                elif (num_h-1)>0:
                    if x == 0:
                        single_coordinate['stack_start_h'] = x*gap_h
                        single_coordinate['stack_end_h'] = x*gap_h+img_h-cut_h
                        single_coordinate['patch_start_h'] = 0
                        single_coordinate['patch_end_h'] = img_h-cut_h
                    elif x == num_h-1:
                        single_coordinate['stack_start_h'] = whole_h-img_h+cut_h
                        single_coordinate['stack_end_h'] = whole_h
                        single_coordinate['patch_start_h'] = cut_h
                        single_coordinate['patch_end_h'] = img_h
                    else:
                        single_coordinate['stack_start_h'] = x*gap_h+cut_h
                        single_coordinate['stack_end_h'] = x*gap_h+img_h-cut_h
                        single_coordinate['patch_start_h'] = cut_h
                        single_coordinate['patch_end_h'] = img_h-cut_h

                if (num_s-1)==0:
                    single_coordinate['stack_start_s'] = 0
                    single_coordinate['stack_end_s'] = whole_s
                    single_coordinate['patch_start_s'] = 0
                    single_coordinate['patch_end_s'] = img_s2
                elif (num_s-1)>0:
                    if z == 0:
                        single_coordinate['stack_start_s'] = z*gap_s2
                        single_coordinate['stack_end_s'] = z*gap_s2+img_s2-cut_s
                        single_coordinate['patch_start_s'] = 0
                        single_coordinate['patch_end_s'] = img_s2-cut_s
                    elif z == num_s-1:
                        single_coordinate['stack_start_s'] = whole_s-img_s2+cut_s
                        single_coordinate['stack_end_s'] = whole_s
                        single_coordinate['patch_start_s'] = cut_s
                        single_coordinate['patch_end_s'] = img_s2
                    else:
                        single_coordinate['stack_start_s'] = z*gap_s2+cut_s
                        single_coordinate['stack_end_s'] = z*gap_s2+img_s2-cut_s
                        single_coordinate['patch_start_s'] = cut_s
                        single_coordinate['patch_end_s'] = img_s2-cut_s

                print('RMBG stack_start_w -----> ',single_coordinate['stack_start_w'], ' patch_start_w -----> ', single_coordinate['patch_start_w'],  \
                'stack_end_w -----> ',single_coordinate['stack_end_w'], ' patch_end_w -----> ',single_coordinate['patch_end_w'])
                print('RMBG stack_start_h -----> ',single_coordinate['stack_start_h'], ' patch_start_h -----> ', single_coordinate['patch_start_h'],  \
                'stack_end_h -----> ',single_coordinate['stack_end_h'], ' patch_end_h -----> ',single_coordinate['patch_end_h'])
                # noise_patch1 = noise_im[init_s:end_s,init_h:end_h,init_w:end_w]
                patch_name = args.datasets_folder+'_x'+str(x)+'_y'+str(y)+'_z'+str(z)
                # train_raw.append(noise_patch1.transpose(1,2,0))
                name_list.append(patch_name)
                # print(' single_coordinate -----> ',single_coordinate)
                coordinate_list[patch_name] = single_coordinate

    return  name_list, noise_im, coordinate_list



def test_preprocess_lessMemoryNoTail_SubImg (args, sub_img):
    img_h = args['img_h']
    img_w = args['img_w']
    img_s2 = args['img_s']
    gap_h = args['gap_h']
    gap_w = args['gap_w']
    gap_s2 = args['gap_s']
    cut_w = (img_w - gap_w)/2
    cut_h = (img_h - gap_h)/2
    cut_s = (img_s2 - gap_s2)/2

    noise_im = sub_img
    noise_im = (noise_im).astype(np.float32)/args['normalize_factor']

    noise_im_ave_single = np.mean(noise_im, axis=0)
    noise_im_ave = np.zeros(noise_im.shape)
    for i in range(0, noise_im.shape[0]):
        noise_im_ave[i,:,:] = noise_im_ave_single
    noise_im = noise_im-noise_im_ave

    whole_w = noise_im.shape[2]
    whole_h = noise_im.shape[1]
    whole_s = noise_im.shape[0]

    num_w = math.ceil((whole_w-img_w+gap_w)/gap_w)
    num_h = math.ceil((whole_h-img_h+gap_h)/gap_h)
    num_s = math.ceil((whole_s-img_s2+gap_s2)/gap_s2)
    # print('int((whole_h-img_h+gap_h)/gap_h) -----> ',int((whole_h-img_h+gap_h)/gap_h))
    # print('int((whole_w-img_w+gap_w)/gap_w) -----> ',int((whole_w-img_w+gap_w)/gap_w))
    # print('int((whole_s-img_s2+gap_s2)/gap_s2) -----> ',int((whole_s-img_s2+gap_s2)/gap_s2))
    name_list = []
    coordinate_list = {}
    for x in range(0,num_h):
        for y in range(0,num_w):
            for z in range(0,num_s):
                single_coordinate={'init_h':0, 'end_h':0, 'init_w':0, 'end_w':0, 'init_s':0, 'end_s':0}
                if x != (num_h-1):
                    init_h = gap_h*x
                    end_h = gap_h*x + img_h
                elif x == (num_h-1):
                    init_h = whole_h - img_h
                    end_h = whole_h

                if y != (num_w-1):
                    init_w = gap_w*y
                    end_w = gap_w*y + img_w
                elif y == (num_w-1):
                    init_w = whole_w - img_w
                    end_w = whole_w

                if z != (num_s-1):
                    init_s = gap_s2*z
                    end_s = gap_s2*z + img_s2
                elif z == (num_s-1):
                    init_s = whole_s - img_s2
                    end_s = whole_s
                single_coordinate['init_h'] = init_h
                single_coordinate['end_h'] = end_h
                single_coordinate['init_w'] = init_w
                single_coordinate['end_w'] = end_w
                single_coordinate['init_s'] = init_s
                single_coordinate['end_s'] = end_s

                if (num_w-1)==0:
                    single_coordinate['stack_start_w'] = 0
                    single_coordinate['stack_end_w'] = whole_w
                    single_coordinate['patch_start_w'] = 0
                    single_coordinate['patch_end_w'] = img_w
                elif (num_w-1)>0:
                    if y == 0:
                        single_coordinate['stack_start_w'] = y*gap_w
                        single_coordinate['stack_end_w'] = y*gap_w+img_w-cut_w
                        single_coordinate['patch_start_w'] = 0
                        single_coordinate['patch_end_w'] = img_w-cut_w
                    elif y == num_w-1:
                        single_coordinate['stack_start_w'] = whole_w-img_w+cut_w
                        single_coordinate['stack_end_w'] = whole_w
                        single_coordinate['patch_start_w'] = cut_w
                        single_coordinate['patch_end_w'] = img_w
                    else:
                        single_coordinate['stack_start_w'] = y*gap_w+cut_w
                        single_coordinate['stack_end_w'] = y*gap_w+img_w-cut_w
                        single_coordinate['patch_start_w'] = cut_w
                        single_coordinate['patch_end_w'] = img_w-cut_w

                if (num_h-1)==0:
                    single_coordinate['stack_start_h'] = 0
                    single_coordinate['stack_end_h'] = whole_h
                    single_coordinate['patch_start_h'] = 0
                    single_coordinate['patch_end_h'] = img_h
                elif (num_h-1)>0:
                    if x == 0:
                        single_coordinate['stack_start_h'] = x*gap_h
                        single_coordinate['stack_end_h'] = x*gap_h+img_h-cut_h
                        single_coordinate['patch_start_h'] = 0
                        single_coordinate['patch_end_h'] = img_h-cut_h
                    elif x == num_h-1:
                        single_coordinate['stack_start_h'] = whole_h-img_h+cut_h
                        single_coordinate['stack_end_h'] = whole_h
                        single_coordinate['patch_start_h'] = cut_h
                        single_coordinate['patch_end_h'] = img_h
                    else:
                        single_coordinate['stack_start_h'] = x*gap_h+cut_h
                        single_coordinate['stack_end_h'] = x*gap_h+img_h-cut_h
                        single_coordinate['patch_start_h'] = cut_h
                        single_coordinate['patch_end_h'] = img_h-cut_h

                if (num_s-1)==0:
                    single_coordinate['stack_start_s'] = 0
                    single_coordinate['stack_end_s'] = whole_s
                    single_coordinate['patch_start_s'] = 0
                    single_coordinate['patch_end_s'] = img_s2
                elif (num_s-1)>0:
                    if z == 0:
                        single_coordinate['stack_start_s'] = z*gap_s2
                        single_coordinate['stack_end_s'] = z*gap_s2+img_s2-cut_s
                        single_coordinate['patch_start_s'] = 0
                        single_coordinate['patch_end_s'] = img_s2-cut_s
                    elif z == num_s-1:
                        single_coordinate['stack_start_s'] = whole_s-img_s2+cut_s
                        single_coordinate['stack_end_s'] = whole_s
                        single_coordinate['patch_start_s'] = cut_s
                        single_coordinate['patch_end_s'] = img_s2
                    else:
                        single_coordinate['stack_start_s'] = z*gap_s2+cut_s
                        single_coordinate['stack_end_s'] = z*gap_s2+img_s2-cut_s
                        single_coordinate['patch_start_s'] = cut_s
                        single_coordinate['patch_end_s'] = img_s2-cut_s
                '''
                print('RMBG stack_start_w -----> ',single_coordinate['stack_start_w'], ' patch_start_w -----> ', single_coordinate['patch_start_w'],  \
                'stack_end_w -----> ',single_coordinate['stack_end_w'], ' patch_end_w -----> ',single_coordinate['patch_end_w'])
                print('RMBG stack_start_h -----> ',single_coordinate['stack_start_h'], ' patch_start_h -----> ', single_coordinate['patch_start_h'],  \
                'stack_end_h -----> ',single_coordinate['stack_end_h'], ' patch_end_h -----> ',single_coordinate['patch_end_h'])
                '''
                # noise_patch1 = noise_im[init_s:end_s,init_h:end_h,init_w:end_w]
                patch_name = 'x'+str(x)+'_y'+str(y)+'_z'+str(z)
                # train_raw.append(noise_patch1.transpose(1,2,0))
                name_list.append(patch_name)
                # print(' single_coordinate -----> ',single_coordinate)
                coordinate_list[patch_name] = single_coordinate

    return  name_list, noise_im, coordinate_list


import cv2
import numpy as np
import scipy.io as scio
# 4邻域的连通域和 8邻域的连通域
# [row, col]
NEIGHBOR_HOODS_4 = True
OFFSETS_4 = [[0, -1], [-1, 0], [0, 0], [1, 0], [0, 1]]

NEIGHBOR_HOODS_8 = False
OFFSETS_8 = [[-1, -1], [0, -1], [1, -1],
             [-1,  0], [0,  0], [1,  0],
             [-1,  1], [0,  1], [1,  1]]



def reorganize(binary_img: np.array):
    index_map = []
    points = []
    index = -1
    rows, cols = binary_img.shape
    for row in range(rows):
        for col in range(cols):
            var = binary_img[row][col]
            if var < 0.5:
                continue
            if var in index_map:
                index = index_map.index(var)
                num = index + 1
            else:
                index = len(index_map)
                num = index + 1
                index_map.append(var)
                points.append([])
            binary_img[row][col] = num
            points[index].append([row, col])
    return binary_img, points



def neighbor_value(binary_img: np.array, offsets, reverse=False):
    rows, cols = binary_img.shape
    label_idx = 0
    rows_ = [0, rows, 1] if reverse == False else [rows-1, -1, -1]
    cols_ = [0, cols, 1] if reverse == False else [cols-1, -1, -1]
    for row in range(rows_[0], rows_[1], rows_[2]):
        for col in range(cols_[0], cols_[1], cols_[2]):
            label = 256
            if binary_img[row][col] < 0.5:
                continue
            for offset in offsets:
                neighbor_row = min(max(0, row+offset[0]), rows-1)
                neighbor_col = min(max(0, col+offset[1]), cols-1)
                neighbor_val = binary_img[neighbor_row, neighbor_col]
                if neighbor_val < 0.5:
                    continue
                label = neighbor_val if neighbor_val < label else label
            if label == 255:
                label_idx += 1
                label = label_idx
            binary_img[row][col] = label
    return binary_img

# binary_img: bg-0, object-255; int
def Two_Pass(binary_img: np.array, neighbor_hoods):
    if neighbor_hoods == 'NEIGHBOR_HOODS_4':
        offsets = [[0, -1], [-1, 0], [0, 0], [1, 0], [0, 1]]
    elif neighbor_hoods == 'NEIGHBOR_HOODS_8':
        offsets = [[-1, -1], [0, -1], [1, -1],
             [-1,  0], [0,  0], [1,  0],
             [-1,  1], [0,  1], [1,  1]]
    else:
        raise ValueError

    binary_img = neighbor_value(binary_img, offsets, False)
    binary_img = neighbor_value(binary_img, offsets, True)

    return binary_img



def recursive_seed(binary_img: np.array, seed_row, seed_col, offsets, num, max_num=100):
    rows, cols = binary_img.shape
    binary_img[seed_row][seed_col] = num
    for offset in offsets:
        neighbor_row = min(max(0, seed_row+offset[0]), rows-1)
        neighbor_col = min(max(0, seed_col+offset[1]), cols-1)
        var = binary_img[neighbor_row][neighbor_col]
        if var < max_num:
            continue
        binary_img = recursive_seed(binary_img, neighbor_row, neighbor_col, offsets, num, max_num)
    return binary_img

# max_num 表示连通域最多存在的个数
def Seed_Filling(binary_img, neighbor_hoods, max_num=1000000):
    if neighbor_hoods == 'NEIGHBOR_HOODS_4':
        offsets = [[0, -1], [-1, 0], [0, 0], [1, 0], [0, 1]]
    elif neighbor_hoods == 'NEIGHBOR_HOODS_8':
        offsets = [[-1, -1], [0, -1], [1, -1],
             [-1,  0], [0,  0], [1,  0],
             [-1,  1], [0,  1], [1,  1]]
    else:
        raise ValueError

    num = 1
    rows, cols = binary_img.shape
    for row in range(rows):
        for col in range(cols):
            var = binary_img[row][col]
            if var <= max_num:
                continue
            binary_img = recursive_seed(binary_img, row, col, offsets, num, max_num=1000000)
            num += 1
    return binary_img


def cal_pccs(x, y, n):
    """
    warning: data format must be narray
    :param x: Variable 1
    :param y: The variable 2
    :param n: The number of elements in x
    :return: pccs
    """
    sum_xy = np.sum(np.sum(x*y))
    sum_x = np.sum(np.sum(x))
    sum_y = np.sum(np.sum(y))
    sum_x2 = np.sum(np.sum(x*x))
    sum_y2 = np.sum(np.sum(y*y))
    pcc = (n*sum_xy-sum_x*sum_y)/np.sqrt((n*sum_x2-sum_x*sum_x)*(n*sum_y2-sum_y*sum_y))
    return pcc


def four_cc_label(img):
    height, width = img.shape
    label = np.zeros((height, width), dtype=np.int32)
    LUT = np.zeros(height * width,dtype=np.uint8)

    COLORS = range(0,1000000)
    out = np.zeros((height, width), dtype=np.uint8)
    label[img[:,:] > 0] = 1

    n = 1
    for y in range(height):
        for x in range(width):
            if y%500==0:
                if x%500==0:
                    print(y,' --- ',height,' --- ',x,' --- ',width)
            if label[y, x] == 0:
                continue
            c2 = label[max(y - 1, 0), min(x + 1, width - 1)]
            c3 = label[max(y - 1, 0), x]
            c4 = label[max(y - 1, 0), max(x - 1, 0)]
            c5 = label[y, max(x - 1, 0)]
            if c3 < 2 and c5 < 2 and c2 < 2 and c4 < 2:
                n += 1
                label[y, x] = n
            else:
                _vs = [c3, c5, c2, c4]
                vs = [a for a in _vs if a > 1]
                v = min(vs)
                label[y, x] = v

                minv = v
                for _v in vs:
                    if LUT[_v] != 0:
                        minv = min(minv, LUT[_v])
                for _v in vs:
                    LUT[_v] = minv

    count = 1
    for l in range(2, n + 1):
        flag = True
        for i in range(n + 1):
            if LUT[i] == l:
                if flag:
                    count += 1
                    flag = False
                LUT[i] = count


    for i, lut in enumerate(LUT[2:]):
        if i%500==0:
            print('LUT',i,len(LUT[2:]))
        out[label == (i + 2)] = COLORS[lut - 2]
    return out



if __name__ == "__main__":
    binary_img = np.zeros((4, 7), dtype=np.int16)
    index = [[0, 2], [0, 5],
            [1, 0], [1, 1], [1, 2], [1, 4], [1, 5], [1, 6],
            [2, 2], [2, 5],
            [3, 1], [3, 2], [3, 4], [3, 6]]
    for i in index:
        binary_img[i[0], i[1]] = np.int16(255)

    print("原始二值图像")
    print(binary_img)

    print("Two_Pass")
    binary_img = Two_Pass(binary_img, NEIGHBOR_HOODS_8)
    binary_img, points = reorganize(binary_img)
    print(binary_img, points)

    print("Seed_Filling")
    binary_img = Seed_Filling(binary_img, NEIGHBOR_HOODS_8)
    binary_img, points = reorganize(binary_img)
    print(binary_img, points)



from setuptools import setup, find_packages
 
setup(
    name="deepwonder_test",
    version="0.2",
    license="MIT Licence",
 
    url="https://github.com/yuanlong-o/Deep_widefield_cal_inferece",
    author="YuanlongZhang&GuoxunZhang",
    author_email="zhanggx19@mails.tsinghua.edu.cn",
 
    packages=find_packages(),
    include_package_data=True,
    platforms="any",
    install_requires=['opencv-python==4.1.2.30',
    'tifffile  ==2021.3.5',
    'scikit-image==0.17.2',
    'scikit-learn==0.24.1',
    'scipy==1.5.2',
    'numpy==1.19.2',
    'torch==1.7.1',
    'torchvision==0.8.2',
    'sklearn==0.0',]
)
